# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import time
import logging
from typing import Optional

from celery import current_app
from bamboo_engine.eri.models import ExecuteInterruptPoint, ScheduleInterruptPoint

from pipeline.eri.celery.queues import QueueResolver

from pipeline.eri.models import Process

logger = logging.getLogger("bamboo_engine")


def _retry_once(action: callable):
    try:
        action()
    except Exception:
        try:
            action()
        except Exception as e:
            raise e


class TaskMixin:
    def _get_task_route_params(self, task_name: str, queue: str, priority: int):
        resolver = QueueResolver(queue)
        queue, routing_key = resolver.resolve_task_queue_and_routing_key(task_name)
        return {
            "queue": queue,
            "priority": priority,
            "routing_key": routing_key,
        }

    def _standardize_headers(self, headers: Optional[dict], process_id: int):
        if headers is None:
            headers = {}

        if "route_info" not in headers:
            process = Process.objects.filter(id=process_id).only("priority", "queue").first()
            headers["route_info"] = {"queue": process.queue, "priority": process.priority}

        headers["timestamp"] = time.time()
        return headers

    def execute(
        self,
        process_id: int,
        node_id: str,
        root_pipeline_id: str,
        parent_pipeline_id: str,
        recover_point: Optional[ExecuteInterruptPoint] = None,
        headers: Optional[dict] = None,
    ):
        """
        派发执行任务，执行任务被拉起执行时应该调用 Engine 实例的 execute 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        """
        task_name = "pipeline.eri.celery.tasks.execute"
        headers = self._standardize_headers(headers=headers, process_id=process_id)
        route_params = self._get_task_route_params(
            task_name=task_name, queue=headers["route_info"]["queue"], priority=headers["route_info"]["priority"]
        )

        def action():
            result = current_app.tasks[task_name].apply_async(
                kwargs={
                    "process_id": process_id,
                    "node_id": node_id,
                    "root_pipeline_id": root_pipeline_id,
                    "parent_pipeline_id": parent_pipeline_id,
                    "recover_point": "{}" if not recover_point else recover_point.to_json(),
                    "headers": headers,
                },
                **route_params,
            )
            logger.info(
                "[pipeline-trace](root_pipeline: %s) node(%s) execute task %s sended",
                root_pipeline_id,
                node_id,
                result.id,
            )

        _retry_once(action=action)

    def schedule(
        self,
        process_id: int,
        node_id: str,
        schedule_id: str,
        callback_data_id: Optional[int] = None,
        recover_point: Optional[ScheduleInterruptPoint] = None,
        headers: Optional[dict] = None,
    ):
        """
        派发调度任务，调度任务被拉起执行时应该调用 Engine 实例的 schedule 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param schedule_id: 调度 ID
        :type schedule_id: str
        """
        task_name = "pipeline.eri.celery.tasks.schedule"
        headers = self._standardize_headers(headers=headers, process_id=process_id)
        route_params = self._get_task_route_params(
            task_name=task_name, queue=headers["route_info"]["queue"], priority=headers["route_info"]["priority"]
        )

        def action():
            result = current_app.tasks[task_name].apply_async(
                kwargs={
                    "process_id": process_id,
                    "node_id": node_id,
                    "schedule_id": schedule_id,
                    "callback_data_id": callback_data_id,
                    "recover_point": "{}" if not recover_point else recover_point.to_json(),
                    "headers": headers,
                },
                **route_params,
            )
            logger.info("[pipeline-trace] node(%s) schedule task %s sended", node_id, result.id)

        _retry_once(action=action)

    def set_next_schedule(
        self,
        process_id: int,
        node_id: str,
        schedule_id: str,
        schedule_after: int,
        headers: Optional[dict] = None,
        callback_data_id: Optional[int] = None,
    ):
        """
        设置下次调度时间，调度倒数归零后应该执行 Engine 实例的 schedule 方法

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param schedule_id: 调度 ID
        :type schedule_id: str
        :param schedule_after: 调度倒数
        :type schedule_after: int
        """
        task_name = "pipeline.eri.celery.tasks.schedule"
        headers = self._standardize_headers(headers=headers, process_id=process_id)
        route_params = self._get_task_route_params(
            task_name=task_name, queue=headers["route_info"]["queue"], priority=headers["route_info"]["priority"]
        )
        headers["timestamp"] += schedule_after

        def action():
            result = current_app.tasks[task_name].apply_async(
                kwargs={
                    "process_id": process_id,
                    "node_id": node_id,
                    "schedule_id": schedule_id,
                    "callback_data_id": callback_data_id,
                    "recover_point": "{}",
                    "headers": headers,
                },
                countdown=schedule_after,
                **route_params,
            )
            logger.info("[pipeline-trace] node(%s) schedule task %s sended", node_id, result.id)

        _retry_once(action=action)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from datetime import timedelta

import timezone_field
from celery import schedules
from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.db import models
from django.db.models import signals
from django.utils.translation import ugettext_lazy as _
from pipeline.contrib.periodic_task.djcelery import managers
from pipeline.contrib.periodic_task.djcelery.tzcrontab import TzAwareCrontab
from pipeline.contrib.periodic_task.djcelery.utils import now
from pipeline.contrib.periodic_task.djcelery.compat import python_2_unicode_compatible


PERIOD_CHOICES = (
    ("days", _("Days")),
    ("hours", _("Hours")),
    ("minutes", _("Minutes")),
    ("seconds", _("Seconds")),
    ("microseconds", _("Microseconds")),
)


@python_2_unicode_compatible
class IntervalSchedule(models.Model):
    every = models.IntegerField(_("every"), null=False)
    period = models.CharField(_("period"), max_length=24, choices=PERIOD_CHOICES,)

    class Meta:
        verbose_name = _("interval")
        verbose_name_plural = _("intervals")
        ordering = ["period", "every"]

    @property
    def schedule(self):
        return schedules.schedule(timedelta(**{self.period: self.every}))

    @classmethod
    def from_schedule(cls, schedule, period="seconds"):
        every = max(schedule.run_every.total_seconds(), 0)
        try:
            return cls.objects.get(every=every, period=period)
        except cls.DoesNotExist:
            return cls(every=every, period=period)
        except MultipleObjectsReturned:
            cls.objects.filter(every=every, period=period).delete()
            return cls(every=every, period=period)

    def __str__(self):
        if self.every == 1:
            return _("every {0.period_singular}").format(self)
        return _("every {0.every:d} {0.period}").format(self)

    @property
    def period_singular(self):
        return self.period[:-1]


def cronexp(field):
    return field and str(field).replace(" ", "") or "*"


@python_2_unicode_compatible
class CrontabSchedule(models.Model):
    minute = models.CharField(_("minute"), max_length=64, default="*")
    hour = models.CharField(_("hour"), max_length=64, default="*")
    day_of_week = models.CharField(_("day of week"), max_length=64, default="*",)
    day_of_month = models.CharField(_("day of month"), max_length=64, default="*",)
    month_of_year = models.CharField(_("month of year"), max_length=64, default="*",)
    timezone = timezone_field.TimeZoneField(default="UTC")

    class Meta:
        verbose_name = _("crontab")
        verbose_name_plural = _("crontabs")
        ordering = ["month_of_year", "day_of_month", "day_of_week", "hour", "minute"]

    def __str__(self):
        return "{} {} {} {} {} (m/h/d/dM/MY)".format(
            cronexp(self.minute),
            cronexp(self.hour),
            cronexp(self.day_of_week),
            cronexp(self.day_of_month),
            cronexp(self.month_of_year),
        )

    @property
    def schedule(self):
        return TzAwareCrontab(
            minute=self.minute,
            hour=self.hour,
            day_of_week=self.day_of_week,
            day_of_month=self.day_of_month,
            month_of_year=self.month_of_year,
            tz=self.timezone,
        )

    @classmethod
    def from_schedule(cls, schedule):
        spec = {
            "minute": schedule._orig_minute,
            "hour": schedule._orig_hour,
            "day_of_week": schedule._orig_day_of_week,
            "day_of_month": schedule._orig_day_of_month,
            "month_of_year": schedule._orig_month_of_year,
            "timezone": schedule.tz,
        }
        try:
            return cls.objects.get(**spec)
        except cls.DoesNotExist:
            return cls(**spec)
        except MultipleObjectsReturned:
            cls.objects.filter(**spec).delete()
            return cls(**spec)


class DjCeleryPeriodicTasks(models.Model):
    ident = models.SmallIntegerField(default=1, primary_key=True, unique=True)
    last_update = models.DateTimeField(null=False)

    objects = managers.ExtendedManager()

    @classmethod
    def changed(cls, instance, **kwargs):
        if not instance.no_changes:
            cls.objects.update_or_create(ident=1, defaults={"last_update": now()})

    @classmethod
    def last_change(cls):
        try:
            return cls.objects.get(ident=1).last_update
        except cls.DoesNotExist:
            pass


@python_2_unicode_compatible
class DjCeleryPeriodicTask(models.Model):
    name = models.CharField(_("name"), max_length=200, unique=True, help_text=_("Useful description"),)
    task = models.CharField(_("task name"), max_length=200)
    interval = models.ForeignKey(
        IntervalSchedule, null=True, blank=True, verbose_name=_("interval"), on_delete=models.CASCADE,
    )
    crontab = models.ForeignKey(
        CrontabSchedule,
        null=True,
        blank=True,
        verbose_name=_("crontab"),
        on_delete=models.CASCADE,
        help_text=_("Use one of interval/crontab"),
    )
    args = models.TextField(_("Arguments"), blank=True, default="[]", help_text=_("JSON encoded positional arguments"),)
    kwargs = models.TextField(
        _("Keyword arguments"), blank=True, default="{}", help_text=_("JSON encoded keyword arguments"),
    )
    queue = models.CharField(
        _("queue"), max_length=200, blank=True, null=True, default=None, help_text=_("Queue defined in CELERY_QUEUES"),
    )
    exchange = models.CharField(_("exchange"), max_length=200, blank=True, null=True, default=None,)
    routing_key = models.CharField(_("routing key"), max_length=200, blank=True, null=True, default=None,)
    expires = models.DateTimeField(_("expires"), blank=True, null=True,)
    enabled = models.BooleanField(_("enabled"), default=True,)
    last_run_at = models.DateTimeField(auto_now=False, auto_now_add=False, editable=False, blank=True, null=True,)
    total_run_count = models.PositiveIntegerField(default=0, editable=False,)
    date_changed = models.DateTimeField(auto_now=True)
    description = models.TextField(_("description"), blank=True)

    objects = managers.PeriodicTaskManager()
    no_changes = False

    class Meta:
        verbose_name = _("djcelery periodic task")
        verbose_name_plural = _("djcelery periodic tasks")

    def validate_unique(self, *args, **kwargs):
        super(DjCeleryPeriodicTask, self).validate_unique(*args, **kwargs)
        if not self.interval and not self.crontab:
            raise ValidationError({"interval": ["One of interval or crontab must be set."]})
        if self.interval and self.crontab:
            raise ValidationError({"crontab": ["Only one of interval or crontab must be set"]})

    def save(self, *args, **kwargs):
        self.exchange = self.exchange or None
        self.routing_key = self.routing_key or None
        self.queue = self.queue or None
        if not self.enabled:
            self.last_run_at = None
        super(DjCeleryPeriodicTask, self).save(*args, **kwargs)

    def __str__(self):
        fmt = "{0.name}: {{no schedule}}"
        if self.interval:
            fmt = "{0.name}: {0.interval}"
        if self.crontab:
            fmt = "{0.name}: {0.crontab}"
        return fmt.format(self)

    @property
    def schedule(self):
        if self.interval:
            return self.interval.schedule
        if self.crontab:
            return self.crontab.schedule


signals.pre_delete.connect(DjCeleryPeriodicTasks.changed, sender=DjCeleryPeriodicTask)
signals.pre_save.connect(DjCeleryPeriodicTasks.changed, sender=DjCeleryPeriodicTask)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy
import re
import logging

from mako.template import Template
from mako import lexer, codegen
from mako.exceptions import MakoException

from pipeline import exceptions
from pipeline.conf.default_settings import MAKO_SAFETY_CHECK
from pipeline.core.data.sandbox import SANDBOX
from pipeline.core.data import mako_safety
from pipeline.utils.mako_utils.checker import check_mako_template_safety
from pipeline.utils.mako_utils.exceptions import ForbiddenMakoTemplateException


logger = logging.getLogger("root")
# find mako template(format is ${xxx}，and ${}# not in xxx, # may raise memory error)
TEMPLATE_PATTERN = re.compile(r"\${[^$#]+}")


def format_constant_key(key):
    """
    @summary: format key to ${key}
    @param key:
    @return:
    """
    return "${%s}" % key


def deformat_constant_key(key):
    """
    @summary: deformat ${key} to key
    @param key:
    @return:
    """
    return key[2:-1]


class ConstantTemplate(object):
    def __init__(self, data):
        self.data = data

    def get_reference(self):
        reference = []
        templates = self.get_templates()
        for tpl in templates:
            reference += self.get_template_reference(tpl)
        reference = list(set(reference))
        return reference

    def get_templates(self):
        templates = []
        data = self.data
        if isinstance(data, str):
            templates += self.get_string_templates(data)
        if isinstance(data, (list, tuple)):
            for item in data:
                templates += ConstantTemplate(item).get_templates()
        if isinstance(data, dict):
            for value in list(data.values()):
                templates += ConstantTemplate(value).get_templates()
        return list(set(templates))

    def resolve_data(self, value_maps):
        data = self.data
        if isinstance(data, str):
            return self.resolve_string(data, value_maps)
        if isinstance(data, list):
            ldata = [""] * len(data)
            for index, item in enumerate(data):
                ldata[index] = ConstantTemplate(copy.deepcopy(item)).resolve_data(value_maps)
            return ldata
        if isinstance(data, tuple):
            ldata = [""] * len(data)
            for index, item in enumerate(data):
                ldata[index] = ConstantTemplate(copy.deepcopy(item)).resolve_data(value_maps)
            return tuple(ldata)
        if isinstance(data, dict):
            for key, value in list(data.items()):
                data[key] = ConstantTemplate(copy.deepcopy(value)).resolve_data(value_maps)
            return data
        return data

    @staticmethod
    def get_string_templates(string):
        return list(set(TEMPLATE_PATTERN.findall(string)))

    @staticmethod
    def get_template_reference(template):
        lex = lexer.Lexer(template)

        try:
            node = lex.parse()
        except MakoException as e:
            logger.warning("pipeline get template[{}] reference error[{}]".format(template, e))
            return []

        # Dummy compiler. _Identifiers class requires one
        # but only interested in the reserved_names field
        def compiler():
            return None

        compiler.reserved_names = set()
        identifiers = codegen._Identifiers(compiler, node)

        return list(identifiers.undeclared)

    @staticmethod
    def resolve_string(string, value_maps):
        if not isinstance(string, str):
            return string
        templates = ConstantTemplate.get_string_templates(string)

        # TODO keep render return object, here only process simple situation
        if len(templates) == 1 and templates[0] == string and deformat_constant_key(string) in value_maps:
            return value_maps[deformat_constant_key(string)]

        for tpl in templates:
            if MAKO_SAFETY_CHECK:
                try:
                    check_mako_template_safety(
                        tpl, mako_safety.SingleLineNodeVisitor(), mako_safety.SingleLinCodeExtractor()
                    )
                except ForbiddenMakoTemplateException as e:
                    logger.warning("forbidden template: {}, exception: {}".format(tpl, e))
                    continue
                except Exception:
                    logger.exception("{} safety check error.".format(tpl))
                    continue
            resolved = ConstantTemplate.resolve_template(tpl, value_maps)
            string = string.replace(tpl, resolved)
        return string

    @staticmethod
    def resolve_template(template, value_maps):
        data = {}
        data.update(SANDBOX)
        data.update(value_maps)
        if not isinstance(template, str):
            raise exceptions.ConstantTypeException("constant resolve error, template[%s] is not a string" % template)
        try:
            tm = Template(template)
        except (MakoException, SyntaxError) as e:
            logger.error("pipeline resolve template[{}] error[{}]".format(template, e))
            return template
        try:
            resolved = tm.render_unicode(**data)
        except Exception as e:
            logger.warning("constant content({}) is invalid, data({}), error: {}".format(template, data, e))
            return template
        else:
            return resolved



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import contextlib
import logging
import traceback

from django.utils.module_loading import import_string

from pipeline.conf import settings
from pipeline.engine.exceptions import InvalidDataBackendError

logger = logging.getLogger("celery")

_backend = None
_candidate_backend = None


def _import_backend(backend_cls_path):
    try:
        backend_cls = import_string(backend_cls_path)
        return backend_cls()
    except ImportError:
        raise InvalidDataBackendError(
            "data backend({}) import error with exception: {}".format(
                settings.PIPELINE_DATA_BACKEND, traceback.format_exc()
            )
        )


@contextlib.contextmanager
def _candidate_exc_ensure(propagate):
    try:
        yield
    except Exception:
        logger.error("candidate data backend operate error: {}".format(traceback.format_exc()))

        if propagate:
            raise


if not _backend:
    _backend = _import_backend(settings.PIPELINE_DATA_BACKEND)

if not _candidate_backend and settings.PIPELINE_DATA_CANDIDATE_BACKEND:
    _candidate_backend = _import_backend(settings.PIPELINE_DATA_CANDIDATE_BACKEND)


if settings.PIPELINE_DATA_BACKEND_AUTO_EXPIRE and not (_backend and _candidate_backend):
    raise RuntimeError(
        "PIPELINE_DATA_BACKEND and PIPELINE_DATA_CANDIDATE_BACKEND can't both be empty when PIPELINE_DATA_BACKEND_AUTO_EXPIRE is set."  # noqa
    )


def _write_operation(method, *args, **kwargs):
    propagate = False

    try:

        if settings.PIPELINE_DATA_BACKEND_AUTO_EXPIRE and method == "set_object":
            # change set_object to expire_cache
            getattr(_backend, "expire_cache")(
                *args, **kwargs, expires=settings.PIPELINE_DATA_BACKEND_AUTO_EXPIRE_SECONDS
            )
        else:
            getattr(_backend, method)(*args, **kwargs)

    except Exception:
        logger.error("data backend operate error: {}".format(traceback.format_exc()))

        if not _candidate_backend:
            raise

        propagate = True

    if _candidate_backend:
        with _candidate_exc_ensure(propagate):
            getattr(_candidate_backend, method)(*args, **kwargs)


def _read_operation(method, *args, **kwargs):
    result = None
    propagate = False

    try:
        result = getattr(_backend, method)(*args, **kwargs)
    except Exception:
        logger.error("data backend operate error: {}".format(traceback.format_exc()))

        if not _candidate_backend:
            raise

        propagate = True

    if result is None and _candidate_backend:
        with _candidate_exc_ensure(propagate):
            result = getattr(_candidate_backend, method)(*args, **kwargs)

    return result


def set_object(key, obj):
    _write_operation("set_object", key, obj)


def del_object(key):
    _write_operation("del_object", key)


def expire_cache(key, obj, expires):
    _write_operation("expire_cache", key, obj, expires)


def get_object(key):
    return _read_operation("get_object", key)


def cache_for(key):
    return _read_operation("cache_for", key)


def set_schedule_data(schedule_id, parent_data):
    return set_object("%s_schedule_parent_data" % schedule_id, parent_data)


def get_schedule_parent_data(schedule_id):
    return get_object("%s_schedule_parent_data" % schedule_id)


def delete_parent_data(schedule_id):
    return del_object("%s_schedule_parent_data" % schedule_id)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from contextlib import contextmanager

import django
from django.db import transaction

if django.VERSION < (1, 6):  # pragma: no cover

    def get_queryset(s):
        return s.get_query_set()


else:

    def get_queryset(s):  # noqa
        return s.get_queryset()


try:
    from django.db.transaction import atomic  # noqa
except ImportError:  # pragma: no cover

    try:
        from django.db.transaction import Transaction  # noqa
    except ImportError:

        @contextmanager
        def commit_on_success(*args, **kwargs):
            try:
                transaction.enter_transaction_management(*args, **kwargs)
                transaction.managed(True, *args, **kwargs)
                try:
                    yield
                except Exception:
                    if transaction.is_dirty(*args, **kwargs):
                        transaction.rollback(*args, **kwargs)
                    raise
                else:
                    if transaction.is_dirty(*args, **kwargs):
                        try:
                            transaction.commit(*args, **kwargs)
                        except Exception:
                            transaction.rollback(*args, **kwargs)
                            raise
            finally:
                transaction.leave_transaction_management(*args, **kwargs)

    else:  # pragma: no cover
        from django.db.transaction import commit_on_success  # noqa

    commit_unless_managed = transaction.commit_unless_managed
    rollback_unless_managed = transaction.rollback_unless_managed
else:

    @contextmanager
    def commit_on_success(using=None):  # noqa
        connection = transaction.get_connection(using)
        if connection.features.autocommits_when_autocommit_is_off:
            # ignore stupid warnings and errors
            yield
        else:
            with transaction.atomic(using):
                yield

    def commit_unless_managed(*args, **kwargs):  # noqa
        pass

    def rollback_unless_managed(*args, **kwargs):  # noqa
        pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline import exceptions
from pipeline.validators.connection import (
    validate_graph_connection,
    validate_graph_without_circle,
)
from pipeline.validators.gateway import validate_gateways, validate_stream
from pipeline.validators.utils import format_pipeline_tree_io_to_list


def validate_pipeline_tree(pipeline_tree, cycle_tolerate=False):
    format_pipeline_tree_io_to_list(pipeline_tree)
    # 1. connection validation
    try:
        validate_graph_connection(pipeline_tree)
    except exceptions.ConnectionValidateError as e:
        raise exceptions.ParserException(e.detail)

    # do not tolerate circle in flow
    if not cycle_tolerate:
        no_cycle = validate_graph_without_circle(pipeline_tree)
        if not no_cycle["result"]:
            raise exceptions.ParserException(no_cycle["message"])

    # 2. gateway validation
    validate_gateways(pipeline_tree)

    # 3. stream validation
    validate_stream(pipeline_tree)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import importlib

from pipeline.conf import settings
from pipeline.constants import PIPELINE_DEFAULT_PRIORITY

adapter_api = importlib.import_module(settings.PIPELINE_ENGINE_ADAPTER_API)


def run_pipeline(pipeline, instance_id=None, check_workers=True, priority=PIPELINE_DEFAULT_PRIORITY, queue=""):
    return adapter_api.run_pipeline(pipeline, instance_id, check_workers=check_workers, priority=priority, queue=queue)


def pause_pipeline(pipeline_id):
    return adapter_api.pause_pipeline(pipeline_id)


def revoke_pipeline(pipeline_id):
    return adapter_api.revoke_pipeline(pipeline_id)


def resume_pipeline(pipeline_id):
    return adapter_api.resume_pipeline(pipeline_id)


def pause_activity(act_id):
    return adapter_api.pause_activity(act_id)


def resume_activity(act_id):
    return adapter_api.resume_activity(act_id)


def retry_activity(act_id, inputs=None):
    return adapter_api.retry_activity(act_id, inputs=inputs)


def skip_activity(act_id):
    return adapter_api.skip_activity(act_id)


def skip_exclusive_gateway(gateway_id, flow_id):
    return adapter_api.skip_exclusive_gateway(gateway_id, flow_id)


def skip_conditional_parallel_gateway(gateway_id, flow_ids, converge_gateway_id):
    return adapter_api.skip_conditional_parallel_gateway(gateway_id, flow_ids, converge_gateway_id)


def forced_fail(act_id, ex_data=""):
    return adapter_api.forced_fail(act_id, ex_data)


def get_state(node_id):
    return adapter_api.get_state(node_id)


def get_topo_tree(pipeline_id):
    return adapter_api.get_topo_tree(pipeline_id)


def get_inputs(act_id):
    return adapter_api.get_inputs(act_id)


def get_outputs(act_id):
    return adapter_api.get_outputs(act_id)


def get_activity_histories(act_id):
    return adapter_api.get_activity_histories(act_id)


def callback(act_id, data=None):
    return adapter_api.callback(act_id, data)


def get_plain_log_for_node(node_id, history_id=-1):
    return adapter_api.get_plain_log_for_node(node_id, history_id)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

import boto3
from botocore.client import Config
from botocore.exceptions import ClientError

from pipeline.contrib.external_plugins.utils.importer.base import (
    AutoInstallRequirementsImporter,
)

logger = logging.getLogger("root")
CONFIG = Config(connect_timeout=10, read_timeout=10, retries={"max_attempts": 2})


class S3ModuleImporter(AutoInstallRequirementsImporter):
    def __init__(
        self,
        name,
        modules,
        service_address,
        bucket,
        access_key,
        secret_key,
        use_cache=True,
        secure_only=True,
        source_dir="",
    ):
        super(S3ModuleImporter, self).__init__(name=name, modules=modules)

        if secure_only and not service_address.startswith("https"):
            raise ValueError("Only accept https when secure_only is True.")
        elif not secure_only:
            logger.warning("Using not secure protocol is extremely dangerous!!")

        self.service_address = (
            service_address
            if service_address.endswith("/")
            else "%s/" % service_address
        )
        # 支持virtual_path模式，bucket形如xxx$virtual_path即判定为该模式
        bucket_partitions = bucket.split("$")
        if len(bucket_partitions) > 1 and bucket_partitions[-1] == "virtual_path":
            self.bucket = "$".join(bucket_partitions[:-1])
            s3_config = Config(
                s3={"addressing_style": "virtual"},
                connect_timeout=10,
                read_timeout=10,
                retries={"max_attempts": 2},
            )
            self.virtual_path = True
        else:
            self.bucket = bucket
            s3_config = CONFIG
            self.virtual_path = False
        self.source_dir = (
            source_dir
            if source_dir == "" or source_dir.endswith("/")
            else "%s/" % source_dir
        )
        self.use_cache = use_cache
        self.s3 = boto3.resource(
            "s3",
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            endpoint_url=self.service_address,
            config=s3_config,
        )
        self.obj_cache = {}

    def is_package(self, fullname):
        return self._fetch_obj_content(self._obj_key(fullname, is_pkg=True)) is not None

    def get_code(self, fullname):
        return compile(self.get_source(fullname), self.get_file(fullname), "exec")

    def get_source(self, fullname):
        source_code = self._fetch_obj_content(
            self._obj_key(fullname, is_pkg=self.is_package(fullname))
        )

        if source_code is None:
            raise ImportError(
                "Can not find {module} in {request_path}/{source_dir}".format(
                    module=fullname,
                    request_path=self.resolve_request_path(),
                    source_dir=self.source_dir,
                )
            )

        return source_code

    def get_path(self, fullname):
        return [self.get_file(fullname).rpartition("/")[0]]

    def resolve_request_path(self):
        if self.virtual_path:
            scheme, host = self.service_address.split("://")
            return f"{scheme}://{self.bucket}.{host}"
        return f"{self.service_address}{self.bucket}"

    def get_file(self, fullname):
        return "{request_path}/{key}".format(
            request_path=self.resolve_request_path(),
            key=self._obj_key(fullname, is_pkg=self.is_package(fullname)),
        )

    def _obj_key(self, fullname, is_pkg):
        base_key = self.source_dir + fullname.replace(".", "/")
        key = "%s/__init__.py" % base_key if is_pkg else "%s.py" % base_key
        return key

    def _fetch_obj_content(self, key):
        logger.info("Try to fetch object: {key}".format(key=key))

        if self.use_cache and key in self.obj_cache:
            logger.info("Use content in cache for s3 object: {key}".format(key=key))
            return self.obj_cache[key]

        obj_content = self._get_s3_obj_content(key)

        if self.use_cache:
            self.obj_cache[key] = obj_content

        return obj_content

    def _get_s3_obj_content(self, key):
        obj = self.s3.Object(bucket_name=self.bucket, key=key)

        try:
            resp = obj.get()
            obj_content = resp["Body"].read()
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                obj_content = None
            else:
                raise

        return obj_content



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
# Generated by Django 1.11.2 on 2017-11-24 10:43


from django.db import migrations, models
import django.db.models.deletion
import pipeline.engine.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Data",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=32, primary_key=True, serialize=False, unique=True, verbose_name="\u8282\u70b9 ID"
                    ),
                ),
                ("inputs", pipeline.engine.models.IOField(verbose_name="\u8f93\u5165\u6570\u636e")),
                ("outputs", pipeline.engine.models.IOField(verbose_name="\u8f93\u51fa\u6570\u636e")),
                ("ex_data", pipeline.engine.models.IOField(verbose_name="\u5f02\u5e38\u6570\u636e")),
            ],
        ),
        migrations.CreateModel(
            name="History",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("identifier", models.CharField(db_index=True, max_length=32, verbose_name="\u8282\u70b9 id")),
                ("started_time", models.DateTimeField(verbose_name="\u5f00\u59cb\u65f6\u95f4")),
                ("archived_time", models.DateTimeField(verbose_name="\u7ed3\u675f\u65f6\u95f4")),
            ],
        ),
        migrations.CreateModel(
            name="HistoryData",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("inputs", pipeline.engine.models.IOField(verbose_name="\u8f93\u5165\u6570\u636e")),
                ("outputs", pipeline.engine.models.IOField(verbose_name="\u8f93\u51fa\u6570\u636e")),
                ("ex_data", pipeline.engine.models.IOField(verbose_name="\u5f02\u5e38\u6570\u636e")),
            ],
        ),
        migrations.CreateModel(
            name="NodeRelationship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("ancestor_id", models.CharField(db_index=True, max_length=32, verbose_name="\u7956\u5148 ID")),
                ("descendant_id", models.CharField(db_index=True, max_length=32, verbose_name="\u540e\u4ee3 ID")),
                ("distance", models.IntegerField(verbose_name="\u8ddd\u79bb")),
            ],
        ),
        migrations.CreateModel(
            name="PipelineModel",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=32, primary_key=True, serialize=False, unique=True, verbose_name="pipeline ID"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PipelineProcess",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=32, primary_key=True, serialize=False, unique=True, verbose_name="Process ID"
                    ),
                ),
                ("root_pipeline_id", models.CharField(max_length=32, verbose_name="\u6839 pipeline \u7684 ID")),
                (
                    "current_node_id",
                    models.CharField(
                        db_index=True,
                        default=b"",
                        max_length=32,
                        verbose_name="\u5f53\u524d\u63a8\u8fdb\u5230\u7684\u8282\u70b9\u7684 ID",
                    ),
                ),
                (
                    "destination_id",
                    models.CharField(
                        default=b"",
                        max_length=32,
                        verbose_name="\u9047\u5230\u8be5 ID \u7684\u8282\u70b9\u5c31\u505c\u6b62\u63a8\u8fdb",
                    ),
                ),
                ("parent_id", models.CharField(default=b"", max_length=32, verbose_name="\u7236 process \u7684 ID")),
                (
                    "ack_num",
                    models.IntegerField(
                        default=0, verbose_name="\u6536\u5230\u5b50\u8282\u70b9 ACK \u7684\u6570\u91cf"
                    ),
                ),
                (
                    "need_ack",
                    models.IntegerField(
                        default=-1,
                        verbose_name="\u9700\u8981\u6536\u5230\u7684\u5b50\u8282\u70b9 ACK \u7684\u6570\u91cf",
                    ),
                ),
                (
                    "is_alive",
                    models.BooleanField(default=True, verbose_name="\u8be5 process \u662f\u5426\u8fd8\u6709\u6548"),
                ),
                (
                    "is_sleep",
                    models.BooleanField(
                        default=False, verbose_name="\u8be5 process \u662f\u5426\u6b63\u5728\u4f11\u7720"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ProcessCeleryTask",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "process_id",
                    models.CharField(
                        db_index=True, max_length=32, unique=True, verbose_name="pipeline \u8fdb\u7a0b ID"
                    ),
                ),
                ("celery_task_id", models.CharField(default=b"", max_length=40, verbose_name="celery \u4efb\u52a1 ID")),
            ],
        ),
        migrations.CreateModel(
            name="ProcessSnapshot",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("data", pipeline.engine.models.IOField(verbose_name="\u5b50 process ID \u4e0e pipeline_stack")),
            ],
        ),
        migrations.CreateModel(
            name="ScheduleService",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=64,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name="ID \u8282\u70b9ID+version",
                    ),
                ),
                ("activity_id", models.CharField(db_index=True, max_length=32, verbose_name="\u8282\u70b9 ID")),
                ("process_id", models.CharField(max_length=32, verbose_name="Pipeline \u8fdb\u7a0b ID")),
                ("schedule_times", models.IntegerField(default=0, verbose_name="\u88ab\u8c03\u5ea6\u6b21\u6570")),
                (
                    "wait_callback",
                    models.BooleanField(default=False, verbose_name="\u662f\u5426\u662f\u56de\u8c03\u578b\u8c03\u5ea6"),
                ),
                (
                    "callback_data",
                    pipeline.engine.models.IOField(default=None, verbose_name="\u56de\u8c03\u6570\u636e"),
                ),
                ("service_act", pipeline.engine.models.IOField(verbose_name="\u5f85\u8c03\u5ea6\u670d\u52a1")),
                ("is_finished", models.BooleanField(default=False, verbose_name="\u662f\u5426\u5df2\u5b8c\u6210")),
                ("version", models.CharField(db_index=True, max_length=32, verbose_name="Activity \u7684\u7248\u672c")),
            ],
        ),
        migrations.CreateModel(
            name="Status",
            fields=[
                (
                    "id",
                    models.CharField(
                        max_length=32, primary_key=True, serialize=False, unique=True, verbose_name="\u8282\u70b9 ID"
                    ),
                ),
                ("state", models.CharField(max_length=10, verbose_name="\u72b6\u6001")),
                ("name", models.CharField(default=b"", max_length=64, verbose_name="\u8282\u70b9\u540d\u79f0")),
                ("retry", models.IntegerField(default=0, verbose_name="\u91cd\u8bd5\u6b21\u6570")),
                ("loop", models.IntegerField(default=1, verbose_name="\u5faa\u73af\u6b21\u6570")),
                ("skip", models.BooleanField(default=False, verbose_name="\u662f\u5426\u8df3\u8fc7")),
                ("created_time", models.DateTimeField(auto_now_add=True, verbose_name="\u521b\u5efa\u65f6\u95f4")),
                ("started_time", models.DateTimeField(null=True, verbose_name="\u5f00\u59cb\u65f6\u95f4")),
                ("archived_time", models.DateTimeField(null=True, verbose_name="\u5f52\u6863\u65f6\u95f4")),
                ("version", models.CharField(max_length=32, verbose_name="\u7248\u672c")),
            ],
            options={"ordering": ["-created_time"]},
        ),
        migrations.CreateModel(
            name="SubProcessRelationship",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("subprocess_id", models.CharField(db_index=True, max_length=32, verbose_name="\u5b50\u6d41\u7a0b ID")),
                ("process_id", models.CharField(max_length=32, verbose_name="\u5bf9\u5e94\u7684\u8fdb\u7a0b ID")),
            ],
        ),
        migrations.AddField(
            model_name="pipelineprocess",
            name="snapshot",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, to="engine.ProcessSnapshot"
            ),
        ),
        migrations.AddField(
            model_name="pipelinemodel",
            name="process",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.SET_NULL, to="engine.PipelineProcess"
            ),
        ),
        migrations.AddField(
            model_name="history",
            name="data",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="engine.HistoryData"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _

from pipeline.component_framework.constants import LEGACY_PLUGINS_VERSION
from pipeline.component_framework.library import ComponentLibrary


class ComponentManager(models.Manager):
    def get_component_dict(self):
        """
        获得标准插件对应的dict类型
        :return:
        """
        components = self.filter(status=True)
        component_dict = {}
        for bundle in components:
            name = bundle.name.split("-")
            group_name = _(name[0])
            name = _(name[1])
            component_dict[bundle.code] = "{}-{}".format(group_name, name)
        return component_dict

    def get_component_dicts(self, other_component_list):
        """
        :param other_component_list: 结果集
        :param index: 结果集中指标字段
        :return:
        """
        components = self.filter(status=True).values("code", "version", "name")
        total = components.count()
        groups = []
        for comp in components:
            version = comp["version"]
            # 插件名国际化
            name = comp["name"].split("-")
            name = "{}-{}-{}".format(_(name[0]), _(name[1]), version)
            code = "{}-{}".format(comp["code"], comp["version"])
            value = 0
            for oth_com_tmp in other_component_list:
                if comp["code"] == oth_com_tmp[1] and comp["version"] == oth_com_tmp[2]:
                    value = oth_com_tmp[0]
            groups.append({"code": code, "name": name, "value": value})
        return total, groups


class ComponentModel(models.Model):
    """
    注册的组件
    """

    code = models.CharField(_("组件编码"), max_length=255, db_index=True)
    version = models.CharField(_("组件版本"), max_length=64, default=LEGACY_PLUGINS_VERSION, db_index=True)
    name = models.CharField(_("组件名称"), max_length=255)
    status = models.BooleanField(_("组件是否可用"), default=True)

    objects = ComponentManager()

    class Meta:
        verbose_name = _("组件 Component")
        verbose_name_plural = _("组件 Component")
        ordering = ["-id"]
        unique_together = (("code", "version"),)

    def __unicode__(self):
        return self.name

    @property
    def group_name(self):
        return ComponentLibrary.get_component_class(self.code, self.version).group_name

    @property
    def group_icon(self):
        return ComponentLibrary.get_component_class(self.code, self.version).group_icon



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import ujson as json

from django.db.models.signals import post_save
from django.dispatch import receiver

from pipeline.component_framework.constants import LEGACY_PLUGINS_VERSION
from pipeline.contrib.statistics.models import (
    ComponentInTemplate,
    TemplateInPipeline,
)
from pipeline.contrib.statistics.tasks import pipeline_post_save_statistics_task, pipeline_archive_statistics_task
from pipeline.contrib.statistics.utils import count_pipeline_tree_nodes
from pipeline.core.constants import PE
from pipeline.models import PipelineInstance, PipelineTemplate
from pipeline.signals import post_pipeline_finish, post_pipeline_revoke

logger = logging.getLogger("root")


@receiver(post_save, sender=PipelineTemplate)
def template_post_save_handler(sender, instance, created, **kwargs):
    """
    模板执行保存处理
    :param sender:
    :param instance: 任务实例 Instance.Object对象
    :param created: 是否是创建（可为更新）
    :param kwargs: 参数序列
    :return:
    """
    template = instance
    template_id = template.template_id
    # 删除原先该项模板数据（无论是更新还是创建，都需要重新创建统计数据）
    ComponentInTemplate.objects.filter(template_id=template_id).delete()
    data = template.data
    component_list = []
    # 任务节点引用标准插件统计（包含间接通过子流程引用）
    for act_id, act in data[PE.activities].items():
        # 标准插件节点直接引用
        if act["type"] == PE.ServiceActivity:
            component = ComponentInTemplate(
                component_code=act["component"]["code"],
                template_id=template_id,
                node_id=act_id,
                version=act["component"].get("version", LEGACY_PLUGINS_VERSION),
            )
            component_list.append(component)
        # 子流程节点间接引用
        else:
            components = ComponentInTemplate.objects.filter(template_id=act["template_id"]).values(
                "subprocess_stack", "component_code", "node_id", "version"
            )
            for component_sub in components:
                # 子流程的执行堆栈（子流程的执行过程）
                stack = json.loads(component_sub["subprocess_stack"])
                # 添加节点id
                stack.insert(0, act_id)
                component = ComponentInTemplate(
                    component_code=component_sub["component_code"],
                    template_id=template_id,
                    node_id=component_sub["node_id"],
                    is_sub=True,
                    subprocess_stack=json.dumps(stack),
                    version=component_sub["version"],
                )
                component_list.append(component)
    ComponentInTemplate.objects.bulk_create(component_list)

    # 统计流程标准插件个数，子流程个数，网关个数
    atom_total, subprocess_total, gateways_total = count_pipeline_tree_nodes(template.data)
    TemplateInPipeline.objects.update_or_create(
        template_id=template_id,
        defaults={"atom_total": atom_total, "subprocess_total": subprocess_total, "gateways_total": gateways_total},
    )


@receiver(post_save, sender=PipelineInstance)
def pipeline_post_save_handler(sender, instance, created, **kwargs):
    try:
        if created:
            pipeline_post_save_statistics_task.delay(instance_id=instance.instance_id)
    except Exception:
        logger.exception("pipeline_post_save_handler[instance_id={}] send message error".format(instance.id))


@receiver(post_pipeline_finish, sender=PipelineInstance)
def pipeline_post_finish_handler(sender, instance_id, **kwargs):
    try:
        pipeline_archive_statistics_task.delay(instance_id=instance_id)
    except Exception:
        logger.exception("pipeline_post_finish_handler[instance_id={}] send message error".format(instance_id))


@receiver(post_pipeline_revoke, sender=PipelineInstance)
def pipeline_post_revoke_handler(sender, instance_id, **kwargs):
    try:
        pipeline_archive_statistics_task.delay(instance_id=instance_id)
    except Exception:
        logger.exception("pipeline_post_revoke_handler[instance_id={}] send message error".format(instance_id))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import os
import traceback

from pipeline.contrib.external_plugins.utils.importer.base import AutoInstallRequirementsImporter

logger = logging.getLogger("root")


class FSModuleImporter(AutoInstallRequirementsImporter):
    def __init__(self, name, modules, path, use_cache=True):
        super(FSModuleImporter, self).__init__(name=name, modules=modules)

        self.path = path if path.endswith("/") else "%s/" % path
        self.use_cache = use_cache
        self.file_cache = {}

    def is_package(self, fullname):
        return os.path.exists(self._file_path(fullname, is_pkg=True))

    def get_code(self, fullname):
        return compile(self.get_source(fullname), self.get_file(fullname), "exec")

    def get_source(self, fullname):
        source_code = self._fetch_file_content(self._file_path(fullname, is_pkg=self.is_package(fullname)))

        if source_code is None:
            raise ImportError("Can not find {module} in {path}".format(module=fullname, path=self.path))

        return source_code

    def get_path(self, fullname):
        return [self._file_path(fullname, is_pkg=True).rpartition("/")[0]]

    def get_file(self, fullname):
        return self._file_path(fullname, is_pkg=self.is_package(fullname))

    def _file_path(self, fullname, is_pkg=False):
        base_path = "{path}{file_path}".format(path=self.path, file_path=fullname.replace(".", "/"))
        file_path = "%s/__init__.py" % base_path if is_pkg else "%s.py" % base_path
        return file_path

    def _fetch_file_content(self, file_path):
        logger.info("Try to fetch file {file_path}".format(file_path=file_path))

        if self.use_cache and file_path in self.file_cache:
            logger.info("Use content in cache for file: {file_path}".format(file_path=file_path))
            return self.file_cache[file_path]

        file_content = self._get_file_content(file_path)

        if self.use_cache:
            self.file_cache[file_path] = file_content

        return file_content

    def _get_file_content(self, file_path):
        try:
            with open(file_path) as f:
                file_content = f.read()
        except IOError:
            logger.info(
                "Error occurred when read {file_path} content: {trace}".format(
                    file_path=file_path, trace=traceback.format_exc()
                )
            )
            file_content = None

        return file_content



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy

import ujson as json

from pipeline import exceptions
from pipeline.utils.collections import FancyDict
from pipeline.utils.utils import convert_bytes_to_str


class DataObject(object):
    def __init__(self, inputs, outputs=None):
        if not isinstance(inputs, dict):
            raise exceptions.DataTypeErrorException("inputs is not dict")
        self.inputs = FancyDict(inputs)
        if outputs is None:
            outputs = {}
        if not isinstance(outputs, dict):
            raise exceptions.DataTypeErrorException("outputs is not dict")
        self.outputs = FancyDict(outputs)

    def get_inputs(self):
        return self.inputs

    def get_outputs(self):
        return self.outputs

    def get_one_of_inputs(self, key, default=None):
        return self.inputs.get(key, default)

    def get_one_of_outputs(self, key, default=None):
        return self.outputs.get(key, default)

    def set_outputs(self, key, value):
        self.outputs.update({key: value})
        return True

    def reset_outputs(self, outputs):
        if not isinstance(outputs, dict):
            raise exceptions.DataTypeErrorException("outputs is not dict")
        self.outputs = FancyDict(outputs)
        return True

    def update_outputs(self, dic):
        self.outputs.update(dic)

    def inputs_copy(self):
        return copy.deepcopy(self.inputs)

    def outputs_copy(self):
        return copy.deepcopy(self.outputs)

    def override_inputs(self, inputs):
        if not isinstance(inputs, FancyDict):
            inputs = FancyDict(inputs)
        self.inputs = inputs

    def override_outputs(self, outputs):
        if not isinstance(outputs, FancyDict):
            outputs = FancyDict(outputs)
        self.outputs = outputs

    def serializer(self):
        result = {"inputs": self.inputs, "outputs": self.outputs}
        return json.dumps(result)

    def __setstate__(self, state):
        # py2 pickle dumps data compatible
        input_key = b"inputs" if b"inputs" in state else "inputs"
        outputs_key = b"outputs" if b"outputs" in state else "outputs"

        self.inputs = FancyDict(convert_bytes_to_str(state[input_key]))
        self.outputs = FancyDict(convert_bytes_to_str(state[outputs_key]))

    def __str__(self):
        return "<inputs: {} | outputs: {}>".format(self.inputs, self.outputs)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from abc import ABCMeta

from pipeline.core.flow.base import FlowNode


def _empty_method(data, parent_data):
    return


class Activity(FlowNode, metaclass=ABCMeta):
    def __init__(self, id, name=None, data=None, failure_handler=None):
        super(Activity, self).__init__(id, name, data)
        self._failure_handler = failure_handler or _empty_method

    def next(self):
        return self.outgoing.unique_one().target

    def failure_handler(self, parent_data):
        return self._failure_handler(data=self.data, parent_data=parent_data)

    def skip(self):
        raise NotImplementedError()

    def prepare_rerun_data(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.constants import PIPELINE_DEFAULT_PRIORITY
from pipeline.engine import api
from pipeline.log.models import LogEntry

STATE_MAP = {
    "CREATED": "RUNNING",
    "READY": "RUNNING",
    "RUNNING": "RUNNING",
    "BLOCKED": "BLOCKED",
    "SUSPENDED": "SUSPENDED",
    "FINISHED": "FINISHED",
    "FAILED": "FAILED",
    "REVOKED": "REVOKED",
}


def run_pipeline(pipeline_instance, instance_id=None, check_workers=True, priority=PIPELINE_DEFAULT_PRIORITY, queue=""):
    return api.start_pipeline(pipeline_instance, check_workers=check_workers, priority=priority, queue=queue)


def pause_pipeline(pipeline_id):
    return api.pause_pipeline(pipeline_id)


def revoke_pipeline(pipeline_id):
    return api.revoke_pipeline(pipeline_id)


def resume_pipeline(pipeline_id):
    return api.resume_pipeline(pipeline_id)


def pause_activity(act_id):
    return api.pause_node_appointment(act_id)


def resume_activity(act_id):
    return api.resume_node_appointment(act_id)


def retry_activity(act_id, inputs=None):
    return api.retry_node(act_id, inputs=inputs)


def skip_activity(act_id):
    return api.skip_node(act_id)


def pause_subprocess(subprocess_id):
    return api.pause_subprocess(subprocess_id)


def skip_exclusive_gateway(gateway_id, flow_id):
    return api.skip_exclusive_gateway(gateway_id, flow_id)


def skip_conditional_parallel_gateway(gateway_id, flow_ids, converge_gateway_id):
    return api.skip_conditional_parallel_gateway(gateway_id, flow_ids, converge_gateway_id)


def forced_fail(node_id, ex_data=""):
    return api.forced_fail(node_id, ex_data=ex_data)


def get_inputs(act_id):
    return api.get_inputs(act_id)


def get_outputs(act_id):
    return api.get_outputs(act_id)


def get_activity_histories(act_id):
    histories = api.get_activity_histories(act_id)
    for item in histories:
        item["started_time"] = _better_time_or_none(item["started_time"])
        item["finished_time"] = _better_time_or_none(item.pop("archived_time"))
    return histories


def callback(act_id, data=None):
    return api.activity_callback(act_id, data)


def get_state(node_id):
    tree = api.get_status_tree(node_id, max_depth=100)

    res = _map(tree)

    # collect all atom
    descendants = {}
    _collect_descendants(tree, descendants)
    res["children"] = descendants

    # return
    return res


def _get_node_state(tree):
    status = []

    # return state when meet leaf
    if not tree.get("children", []):
        return STATE_MAP[tree["state"]]

    # iterate children and get child state recursively
    for identifier_code, child_tree in list(tree["children"].items()):
        status.append(_get_node_state(child_tree))

    # summary parent state
    return STATE_MAP[_get_parent_state_from_children_state(tree["state"], status)]


def _get_parent_state_from_children_state(parent_state, children_state_list):
    """
    @summary: 根据子任务状态计算父任务状态
    @param parent_state:
    @param children_state_list:
    @return:
    """
    children_state_set = set(children_state_list)
    if parent_state == "BLOCKED":
        if "RUNNING" in children_state_set:
            parent_state = "RUNNING"
        if "FAILED" in children_state_set:
            parent_state = "FAILED"
    return parent_state


def _collect_descendants(tree, descendants):
    # iterate children for tree
    for identifier_code, child_tree in list(tree["children"].items()):
        child_status = _map(child_tree)
        descendants[identifier_code] = child_status

        # collect children
        if child_tree["children"]:
            _collect_descendants(child_tree, descendants)


def _better_time_or_none(time):
    return time.strftime("%Y-%m-%d %H:%M:%S") if time else time


def _map(tree):
    tree.setdefault("children", {})
    return {
        "id": tree["id"],
        "state": _get_node_state(tree),
        "start_time": _better_time_or_none(tree["started_time"]),
        "finish_time": _better_time_or_none(tree["archived_time"]),
        "loop": tree["loop"],
        "retry": tree["retry"],
        "skip": tree["skip"],
    }


def get_plain_log_for_node(node_id, history_id):
    return LogEntry.objects.plain_log_for_node(node_id=node_id, history_id=history_id)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import importlib
import logging
import traceback

from pipeline.contrib.external_plugins.models import source_cls_factory
from pipeline.contrib.external_plugins.utils.importer import importer_context

logger = logging.getLogger("root")


def load_external_modules():
    for source_type, source_model_cls in list(source_cls_factory.items()):
        # get all external source
        sources = source_model_cls.objects.all()

        # get importer for source
        for source in sources:
            _import_modules_in_source(source)


def _import_modules_in_source(source):
    try:
        importer = source.importer()

        with importer_context(importer):
            for mod in source.modules:
                importlib.import_module(mod)
    except Exception:
        logger.error("An error occurred when loading {{{}}}: {}".format(source.name, traceback.format_exc()))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from pipeline.core.data import var
from pipeline.core.flow.event import EmptyStartEvent
from pipeline.engine.models import Status

from .base import FlowElementHandler

logger = logging.getLogger("pipeline_engine")

__all__ = ["EmptyStartEventHandler"]


class EmptyStartEventHandler(FlowElementHandler):
    @staticmethod
    def element_cls():
        return EmptyStartEvent

    @staticmethod
    def _hydrate(value):
        return value.get() if issubclass(value.__class__, var.Variable) else value

    def handle(self, process, element, status):
        # 进行变量预渲染
        if hasattr(element.data, "inputs"):
            for pre_render_key in element.data.inputs.get("pre_render_keys", []):
                context_variables = process.top_pipeline.context.variables
                if pre_render_key in context_variables:
                    context_variables[pre_render_key] = self._hydrate(context_variables[pre_render_key])

        Status.objects.finish(element)
        return self.HandleResult(next_node=element.next(), should_return=False, should_sleep=False)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy
import hashlib
import logging
import queue
import zlib

import ujson as json
from django.db import models, transaction
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from pipeline.conf import settings
from pipeline.constants import PIPELINE_DEFAULT_PRIORITY
from pipeline.core.constants import PE
from pipeline.signals import post_pipeline_finish, post_pipeline_revoke
from pipeline.engine.utils import ActionResult, calculate_elapsed_time
from pipeline.exceptions import SubprocessRefError
from pipeline.parser.context import get_pipeline_context
from pipeline.parser.utils import replace_all_id
from pipeline.service import task_service
from pipeline.utils.graph import Graph
from pipeline.utils.uniqid import node_uniqid, uniqid

MAX_LEN_OF_NAME = 128
logger = logging.getLogger("root")


class CompressJSONField(models.BinaryField):
    def __init__(self, compress_level=6, *args, **kwargs):
        super(CompressJSONField, self).__init__(*args, **kwargs)
        self.compress_level = compress_level

    def get_prep_value(self, value):
        value = super(CompressJSONField, self).get_prep_value(value)
        return zlib.compress(json.dumps(value).encode("utf-8"), self.compress_level)

    def to_python(self, value):
        value = super(CompressJSONField, self).to_python(value)
        return json.loads(zlib.decompress(value).decode("utf-8"))

    def from_db_value(self, value, expression, connection, context=None):
        return self.to_python(value)


class SnapshotManager(models.Manager):
    def create_snapshot(self, data):
        h = hashlib.md5()
        h.update(json.dumps(data).encode("utf-8"))
        snapshot = self.create(md5sum=h.hexdigest(), data=data)
        return snapshot

    def data_for_snapshot(self, snapshot_id):
        return self.get(id=snapshot_id).data


class Snapshot(models.Model):
    """
    数据快照
    """

    md5sum = models.CharField(_("快照字符串的md5sum"), max_length=32, db_index=True)
    create_time = models.DateTimeField(_("创建时间"), auto_now_add=True)
    data = CompressJSONField(null=True, blank=True)

    objects = SnapshotManager()

    class Meta:
        verbose_name = _("模板快照")
        verbose_name_plural = _("模板快照")
        ordering = ["-id"]
        app_label = "pipeline"

    def __unicode__(self):
        return str(self.md5sum)

    def has_change(self, data):
        """
        检测 data 的 md5 是否和当前存储的不一致
        @param data:
        @return: 新的 md5，md5 是否有变化
        """
        h = hashlib.md5()
        h.update(json.dumps(data).encode("utf-8"))
        md5 = h.hexdigest()
        return md5, self.md5sum != md5


class TreeInfo(models.Model):
    """
    pipeline 数据信息
    """

    data = CompressJSONField(null=True, blank=True)


def get_subprocess_act_list(pipeline_data):
    """
    获取 pipeline 结构中所有的子流程节点
    @param pipeline_data: 流程结构数据
    @return: 子流程节点
    """
    activities = pipeline_data[PE.activities]
    act_ids = [act_id for act_id in activities if activities[act_id][PE.type] == PE.SubProcess]
    return [activities[act_id] for act_id in act_ids]


def _act_id_in_graph(act):
    """
    获取子流程节点引用的模板 ID
    @param act: 子流程节点
    @return: 模板 ID:版本 或 模板ID
    """
    return "{}:{}".format(act["template_id"], act["version"]) if act.get("version") else act["template_id"]


class TemplateManager(models.Manager):
    def subprocess_ref_validate(self, data, root_id=None, root_name=None):
        """
        验证子流程引用是否合法
        @param data:
        @param root_id:
        @param root_name:
        @return: 引用是否合法，相关信息
        """
        try:
            sub_refs, name_map = self.construct_subprocess_ref_graph(data, root_id=root_id, root_name=root_name)
        except PipelineTemplate.DoesNotExist as e:
            return False, str(e)

        nodes = list(sub_refs.keys())
        flows = []
        for node in nodes:
            for ref in sub_refs[node]:
                if ref in nodes:
                    flows.append([node, ref])
        graph = Graph(nodes, flows)
        # circle reference check
        trace = graph.get_cycle()
        if trace:
            name_trace = " → ".join([name_map[proc_id] for proc_id in trace])
            return False, _("子流程引用链中存在循环引用：%s") % name_trace

        return True, ""

    def create_model(self, structure_data, **kwargs):
        """
        创建流程模板对象
        @param structure_data: pipeline 结构数据
        @param kwargs: 其他参数
        @return: 流程模板
        """
        result, msg = self.subprocess_ref_validate(structure_data)

        if not result:
            raise SubprocessRefError(msg)

        snapshot = Snapshot.objects.create_snapshot(structure_data)
        kwargs["snapshot"] = snapshot
        kwargs["template_id"] = node_uniqid()
        obj = self.create(**kwargs)
        # version track
        # TemplateVersion.objects.track(obj)

        return obj

    def delete_model(self, template_ids):
        """
        删除模板对象
        @param template_ids: 模板对象 ID 列表或 ID
        @return:
        """
        if not isinstance(template_ids, list):
            template_ids = [template_ids]
        qs = self.filter(template_id__in=template_ids)
        for template in qs:
            template.is_deleted = True
            template.name = uniqid()
            template.save()

    def construct_subprocess_ref_graph(self, pipeline_data, root_id=None, root_name=None):
        """
        构造子流程引用图
        @param pipeline_data: pipeline 结构数据
        @param root_id: 所有引用开始的根流程 ID
        @param root_name: 根流程名
        @return: 子流程引用图，模板 ID -> 模板姓名映射字典
        """
        subprocess_act = get_subprocess_act_list(pipeline_data)
        tid_queue = queue.Queue()
        graph = {}
        version = {}
        name_map = {}

        if root_id:
            graph[root_id] = [_act_id_in_graph(act) for act in subprocess_act]
            name_map[root_id] = root_name

        for act in subprocess_act:
            tid_queue.put(_act_id_in_graph(act))
            version[_act_id_in_graph(act)] = act.get("version")

        while not tid_queue.empty():
            tid = tid_queue.get()
            template = self.get(template_id=tid.split(":")[0])
            name_map[tid] = template.name
            subprocess_act = get_subprocess_act_list(template.data_for_version(version[tid]))

            for act in subprocess_act:
                ref_tid = _act_id_in_graph(act)
                graph.setdefault(tid, []).append(ref_tid)
                version[_act_id_in_graph(act)] = act.get("version")
                if ref_tid not in graph:
                    tid_queue.put(ref_tid)
            if not subprocess_act:
                graph[tid] = []

        return graph, name_map

    def unfold_subprocess(self, pipeline_data):
        """
        展开 pipeline 数据中所有的子流程
        @param pipeline_data: pipeline 数据
        @return:
        """
        id_maps = replace_all_id(pipeline_data)
        activities = pipeline_data[PE.activities]
        for act_id, act in list(activities.items()):
            if act[PE.type] == PE.SubProcess:
                subproc_data = self.get(template_id=act[PE.template_id]).data_for_version(act.get(PE.version))

                sub_id_maps = self.unfold_subprocess(subproc_data)
                # act_id is new id
                id_maps[PE.subprocess_detail].update({act_id: sub_id_maps})

                subproc_data[PE.id] = act_id
                act[PE.pipeline] = subproc_data
        return id_maps

    def replace_id(self, pipeline_data):
        """
        替换 pipeline 中所有 ID
        @param pipeline_data: pipeline 数据
        @return:
        """
        id_maps = replace_all_id(pipeline_data)
        activities = pipeline_data[PE.activities]
        for act_id, act in list(activities.items()):
            if act[PE.type] == PE.SubProcess:
                subproc_data = act[PE.pipeline]
                sub_id_maps = self.replace_id(subproc_data)
                # act_id is new id
                id_maps[PE.subprocess_detail].update({act_id: sub_id_maps})

                subproc_data[PE.id] = act_id
                act[PE.pipeline] = subproc_data
        return id_maps


class PipelineTemplate(models.Model):
    """
    流程模板
    """

    template_id = models.CharField(_("模板ID"), max_length=32, unique=True)
    name = models.CharField(_("模板名称"), max_length=MAX_LEN_OF_NAME, default="default_template", db_index=True)
    create_time = models.DateTimeField(_("创建时间"), auto_now_add=True, db_index=True)
    creator = models.CharField(_("创建者"), max_length=32)
    description = models.TextField(_("描述"), null=True, blank=True)
    editor = models.CharField(_("修改者"), max_length=32, null=True, blank=True)
    edit_time = models.DateTimeField(_("修改时间"), auto_now=True, db_index=True)
    snapshot = models.ForeignKey(
        Snapshot, verbose_name=_("模板结构数据"), related_name="snapshot_templates", on_delete=models.DO_NOTHING
    )
    has_subprocess = models.BooleanField(_("是否含有子流程"), default=False)
    is_deleted = models.BooleanField(_("是否删除"), default=False, help_text=_("表示当前模板是否删除"))

    objects = TemplateManager()

    class Meta:
        verbose_name = _("Pipeline模板")
        verbose_name_plural = _("Pipeline模板")
        ordering = ["-edit_time"]
        app_label = "pipeline"

    def __unicode__(self):
        return "{}-{}".format(self.template_id, self.name)

    @property
    def data(self):
        return self.snapshot.data

    @property
    def version(self):
        return self.snapshot.md5sum

    @property
    def subprocess_version_info(self):
        # 1. get all subprocess
        subprocess_info = TemplateRelationship.objects.get_subprocess_info(self.template_id).values(
            "descendant_template_id", "subprocess_node_id", "version", "always_use_latest"
        )
        info = {"subproc_has_update": False, "details": []}
        if not subprocess_info:
            return info

        # 2. check whether subprocess is expired
        temp_current_versions = {
            item.template_id: item
            for item in TemplateCurrentVersion.objects.filter(
                template_id__in=[item["descendant_template_id"] for item in subprocess_info]
            )
        }

        expireds = []
        for item in subprocess_info:
            item["expired"] = (
                False
                if item["version"] is None
                or item["descendant_template_id"] not in temp_current_versions
                or item["always_use_latest"]
                else (item["version"] != temp_current_versions[item["descendant_template_id"]].current_version)
            )
            info["details"].append(item)
            expireds.append(item["expired"])

        info["subproc_has_update"] = any(expireds)

        # 3. return
        return info

    @property
    def subprocess_has_update(self):
        return self.subprocess_version_info["subproc_has_update"]

    def data_for_version(self, version):
        """
        获取某个版本的模板数据
        @param version: 版本号
        @return: 模板数据
        """
        if not version:
            return self.data
        return Snapshot.objects.filter(md5sum=version).order_by("-id").first().data

    def referencer(self):
        """
        获取引用了该模板的其他模板
        @return: 引用了该模板的其他模板 ID 列表
        """
        referencer = TemplateRelationship.objects.referencer(self.template_id)
        template_id = self.__class__.objects.filter(template_id__in=referencer, is_deleted=False).values_list(
            "template_id", flat=True
        )
        return list(template_id)

    def clone_data(self):
        """
        获取该模板数据的克隆
        @return: ID 替换过后的模板数据
        """
        data = self.data
        replace_all_id(self.data)
        return data

    def update_template(self, structure_data, **kwargs):
        """
        更新当前模板的模板数据
        @param structure_data: pipeline 结构数据
        @param kwargs: 其他参数
        @return:
        """
        result, msg = PipelineTemplate.objects.subprocess_ref_validate(structure_data, self.template_id, self.name)
        if not result:
            raise SubprocessRefError(msg)

        snapshot = Snapshot.objects.create_snapshot(structure_data)
        kwargs["snapshot"] = snapshot
        kwargs["edit_time"] = timezone.now()
        exclude_keys = ["template_id", "creator", "create_time", "is_deleted"]
        for key in exclude_keys:
            kwargs.pop(key, None)
        for key, value in list(kwargs.items()):
            setattr(self, key, value)
        self.save()

    def gen_instance(self, inputs=None, **kwargs):
        """
        使用该模板创建实例
        @param inputs: 自定义输入
        @param kwargs: 其他参数
        @return: 实例对象
        """
        instance, _ = PipelineInstance.objects.create_instance(
            template=self, exec_data=copy.deepcopy(self.data), inputs=inputs, **kwargs
        )
        return instance

    def set_has_subprocess_bit(self):
        acts = list(self.data[PE.activities].values())
        self.has_subprocess = any([act["type"] == PE.SubProcess for act in acts])


class TemplateRelationShipManager(models.Manager):
    def get_subprocess_info(self, template_id):
        """
        获取某个模板中所有的子流程信息
        @param template_id: 模板 ID
        @return: 该模板所引用的子流程相关信息
        """
        return self.filter(ancestor_template_id=template_id)

    def referencer(self, template_id):
        """
        获取引用了某个模板的其他模板
        @param template_id: 被引用的模板
        @return: 引用了该模板的其他模板 ID 列表
        """
        return list(set(self.filter(descendant_template_id=template_id).values_list("ancestor_template_id", flat=True)))


class TemplateRelationship(models.Model):
    """
    流程模板引用关系：直接引用
    """

    ancestor_template_id = models.CharField(_("根模板ID"), max_length=32, db_index=True)
    descendant_template_id = models.CharField(_("子流程模板ID"), max_length=32, null=False, db_index=True)
    subprocess_node_id = models.CharField(_("子流程节点 ID"), max_length=32, null=False)
    version = models.CharField(_("快照字符串的md5"), max_length=32, null=False)
    always_use_latest = models.BooleanField(_("是否永远使用最新版本"), default=False)

    objects = TemplateRelationShipManager()


class TemplateCurrentVersionManager(models.Manager):
    def update_current_version(self, template):
        """
        更新某个模板的当前版本
        @param template: 模板对象
        @return: 记录模板当前版本的对象
        """
        obj, __ = self.update_or_create(
            template_id=template.template_id, defaults={"current_version": template.version}
        )
        return obj


class TemplateCurrentVersion(models.Model):
    """
    记录流程模板当前版本的表
    """

    template_id = models.CharField(_("模板ID"), max_length=32, db_index=True)
    current_version = models.CharField(_("快照字符串的md5"), max_length=32, null=False)

    objects = TemplateCurrentVersionManager()


class TemplateVersionManager(models.Manager):
    def track(self, template):
        """
        记录模板的版本号
        @param template: 被记录模板
        @return: 版本跟踪对象
        """
        if not template.snapshot:
            return None

        # don't track if latest version is same as current version
        versions = self.filter(template_id=template.id).order_by("-id")
        if versions and versions[0].md5 == template.snapshot.md5sum:
            return versions[0]

        return self.create(template=template, snapshot=template.snapshot, md5=template.snapshot.md5sum)


class TemplateVersion(models.Model):
    """
    模板版本号记录节点
    """

    template = models.ForeignKey(PipelineTemplate, verbose_name=_("模板 ID"), null=False, on_delete=models.CASCADE)
    snapshot = models.ForeignKey(Snapshot, verbose_name=_("模板数据 ID"), null=False, on_delete=models.CASCADE)
    md5 = models.CharField(_("快照字符串的md5"), max_length=32, db_index=True)
    date = models.DateTimeField(_("添加日期"), auto_now_add=True)

    objects = TemplateVersionManager()


class TemplateScheme(models.Model):
    """
    模板执行方案
    """

    template = models.ForeignKey(
        PipelineTemplate, verbose_name=_("对应模板 ID"), null=False, blank=False, on_delete=models.CASCADE
    )
    unique_id = models.CharField(_("方案唯一ID"), max_length=97, unique=True, null=False, blank=True)
    name = models.CharField(_("方案名称"), max_length=64, null=False, blank=False)
    edit_time = models.DateTimeField(_("修改时间"), auto_now=True)
    data = CompressJSONField(verbose_name=_("方案数据"))

    subprocess_scheme_relation = models.ManyToManyField(verbose_name=_("子流程节点引用执行方案的关系"), to=TemplateRelationship)


class InstanceManager(models.Manager):
    def create_instance(self, template, exec_data, spread=False, inputs=None, **kwargs):
        """
        创建流程实例对象
        @param template: 流程模板
        @param exec_data: 执行用流程数据
        @param spread: exec_data 是否已经展开
        @param kwargs: 其他参数
        @param inputs: 自定义输入
        @return: 实例对象
        """
        if not spread:
            id_maps = PipelineTemplate.objects.unfold_subprocess(exec_data)
        else:
            id_maps = PipelineTemplate.objects.replace_id(exec_data)

        inputs = inputs or {}

        for key, val in list(inputs.items()):
            if key in exec_data["data"]["inputs"]:
                exec_data["data"]["inputs"][key]["value"] = val

        instance_id = node_uniqid()
        exec_data["id"] = instance_id
        exec_snapshot = Snapshot.objects.create_snapshot(exec_data)
        TreeInfo.objects.create()
        if template is not None:
            kwargs["template"] = template
            kwargs["snapshot_id"] = template.snapshot.id
        kwargs["instance_id"] = instance_id
        kwargs["execution_snapshot_id"] = exec_snapshot.id
        return self.create(**kwargs), id_maps

    def delete_model(self, instance_ids):
        """
        删除流程实例对象
        @param instance_ids: 实例 ID 或 ID 列表
        @return:
        """
        if not isinstance(instance_ids, list):
            instance_ids = [instance_ids]
        qs = self.filter(instance_id__in=instance_ids)
        for instance in qs:
            instance.is_deleted = True
            instance.name = uniqid()
            instance.save()

    def set_started(self, instance_id, executor):
        """
        将实例的状态设置为已开始
        @param instance_id: 实例 ID
        @param executor: 执行者
        @return:
        """
        self.filter(instance_id=instance_id).update(start_time=timezone.now(), is_started=True, executor=executor)

    def set_finished(self, instance_id):
        """
        将实例的状态设置为已完成
        @param instance_id: 实例 ID
        @return:
        """
        self.filter(instance_id=instance_id).update(finish_time=timezone.now(), is_finished=True)
        post_pipeline_finish.send(sender=PipelineInstance, instance_id=instance_id)

    def set_revoked(self, instance_id):
        """
        将实例的状态设置为已撤销
        @param instance_id: 实例 ID
        @return:
        """
        self.filter(instance_id=instance_id).update(finish_time=timezone.now(), is_revoked=True)
        post_pipeline_revoke.send(sender=PipelineInstance, instance_id=instance_id)


class PipelineInstance(models.Model):
    """
    流程实例对象
    """

    instance_id = models.CharField(_("实例ID"), max_length=32, unique=True, db_index=True)
    template = models.ForeignKey(
        PipelineTemplate, verbose_name=_("Pipeline模板"), null=True, blank=True, on_delete=models.SET_NULL
    )
    name = models.CharField(_("实例名称"), max_length=MAX_LEN_OF_NAME, default="default_instance")
    creator = models.CharField(_("创建者"), max_length=32, blank=True)
    create_time = models.DateTimeField(_("创建时间"), auto_now_add=True, db_index=True)
    executor = models.CharField(_("执行者"), max_length=32, blank=True)
    start_time = models.DateTimeField(_("启动时间"), null=True, blank=True)
    finish_time = models.DateTimeField(_("结束时间"), null=True, blank=True)
    description = models.TextField(_("描述"), blank=True)
    is_started = models.BooleanField(_("是否已经启动"), default=False)
    is_finished = models.BooleanField(_("是否已经完成"), default=False)
    is_revoked = models.BooleanField(_("是否已经撤销"), default=False)
    is_deleted = models.BooleanField(_("是否已经删除"), default=False, help_text=_("表示当前实例是否删除"))
    is_expired = models.BooleanField(_("是否已经过期"), default=False, help_text=_("运行时被定期清理即为过期"))
    snapshot = models.ForeignKey(
        Snapshot,
        blank=True,
        null=True,
        related_name="snapshot_instances",
        verbose_name=_("实例结构数据，指向实例对应的模板的结构数据"),
        on_delete=models.SET_NULL,
    )
    execution_snapshot = models.ForeignKey(
        Snapshot,
        blank=True,
        null=True,
        related_name="execution_snapshot_instances",
        verbose_name=_("用于实例执行的结构数据"),
        on_delete=models.SET_NULL,
    )
    tree_info = models.ForeignKey(
        TreeInfo,
        blank=True,
        null=True,
        related_name="tree_info_instances",
        verbose_name=_("提前计算好的一些流程结构数据"),
        on_delete=models.SET_NULL,
    )

    objects = InstanceManager()

    class Meta:
        verbose_name = _("Pipeline实例")
        verbose_name_plural = _("Pipeline实例")
        ordering = ["-create_time"]
        app_label = "pipeline"

    def __unicode__(self):
        return "{}-{}".format(self.instance_id, self.name)

    @property
    def data(self):
        return self.snapshot.data

    @property
    def execution_data(self):
        return self.execution_snapshot.data

    @property
    def node_id_set(self):
        if not self.tree_info:
            self.calculate_tree_info(save=True)
        return set(self.tree_info.data["node_id_set"])

    @property
    def elapsed_time(self):
        return calculate_elapsed_time(self.start_time, self.finish_time)

    def set_execution_data(self, data):
        """
        设置实例的执行用流程数据
        @param data: 执行用流程数据
        @return:
        """
        self.execution_snapshot.data = data
        self.execution_snapshot.save()

    def _replace_id(self, exec_data):
        """
        替换执行用流程数据中的所有 ID
        @param exec_data: 执行用流程数据
        @return:
        """
        replace_all_id(exec_data)
        activities = exec_data[PE.activities]
        for act_id, act in list(activities.items()):
            if act[PE.type] == PE.SubProcess:
                self._replace_id(act["pipeline"])
                act["pipeline"]["id"] = act_id

    def clone(self, creator, **kwargs):
        """
        返回当前实例对象的克隆
        @param creator: 创建者
        @param kwargs: 其他参数
        @return: 当前实例对象的克隆
        """
        name = kwargs.get("name") or timezone.localtime(timezone.now()).strftime("clone%Y%m%d%H%m%S")
        instance_id = node_uniqid()

        exec_data = self.execution_data
        self._replace_id(exec_data)
        # replace root id
        exec_data["id"] = instance_id
        new_snapshot = Snapshot.objects.create_snapshot(exec_data)

        return self.__class__.objects.create(
            template=self.template,
            instance_id=instance_id,
            name=name,
            creator=creator,
            description=self.description,
            snapshot=self.snapshot,
            execution_snapshot=new_snapshot,
        )

    def start(self, executor, check_workers=True, priority=PIPELINE_DEFAULT_PRIORITY, queue=""):
        """
        启动当前流程
        @param executor: 执行者
        @param check_workers: 是否检测 worker 的状态
        @return: 执行结果
        """

        with transaction.atomic():
            instance = self.__class__.objects.select_for_update().get(id=self.id)
            if instance.is_started:
                return ActionResult(result=False, message="pipeline instance already started.")

            pipeline_data = instance.execution_data

            try:
                parser_cls = import_string(settings.PIPELINE_PARSER_CLASS)
            except ImportError:
                return ActionResult(result=False, message="invalid parser class: %s" % settings.PIPELINE_PARSER_CLASS)

            instance.start_time = timezone.now()
            instance.is_started = True
            instance.executor = executor

            parser = parser_cls(pipeline_data)
            pipeline = parser.parse(
                root_pipeline_data=get_pipeline_context(
                    instance, obj_type="instance", data_type="data", username=executor
                ),
                root_pipeline_context=get_pipeline_context(
                    instance, obj_type="instance", data_type="context", username=executor
                ),
            )

            # calculate tree info
            instance.calculate_tree_info()

            instance.save()

        act_result = task_service.run_pipeline(pipeline, check_workers=check_workers, priority=priority, queue=queue)

        if not act_result.result:
            with transaction.atomic():
                instance = self.__class__.objects.select_for_update().get(id=self.id)
                instance.start_time = None
                instance.is_started = False
                instance.executor = ""
                instance.save()

        return act_result

    def _get_node_id_set(self, node_id_set, data):
        """
        递归获取当前实例中所有节点的 ID（包括子流程中的节点）
        @param node_id_set: 节点 ID 集合
        @param data: 流程数据
        @return:
        """
        node_id_set.add(data[PE.start_event]["id"])
        node_id_set.add(data[PE.end_event]["id"])
        for gid in data[PE.gateways]:
            node_id_set.add(gid)
        for aid, act_data in list(data[PE.activities].items()):
            node_id_set.add(aid)
            if act_data[PE.type] == PE.SubProcess:
                self._get_node_id_set(node_id_set, act_data["pipeline"])

    def calculate_tree_info(self, save=False):
        """
        计算当前流程实例执行用流程数据中的一些基本信息
        @param save: 是否在计算完后保存实例对象
        @return:
        """
        self.tree_info = TreeInfo.objects.create()
        node_id_set = set({})

        # get node id set
        self._get_node_id_set(node_id_set, self.execution_data)

        tree_info = {"node_id_set": list(node_id_set)}
        self.tree_info.data = tree_info
        self.tree_info.save()

        if save:
            self.save()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import sys
import traceback

from django.apps import AppConfig
from django.conf import settings
from django.db.utils import ProgrammingError

from pipeline.utils import env

logger = logging.getLogger("root")

DJANGO_MANAGE_CMD = "manage.py"
DEFAULT_TRIGGERS = {"runserver", "celery", "worker", "uwsgi", "shell", "update_component_models"}


class ExternalPluginsConfig(AppConfig):
    name = "pipeline.contrib.external_plugins"
    label = "pipeline_external_plugins"
    verbose_name = "PipelineExternalPlugins"

    def ready(self):
        from pipeline.contrib.external_plugins import loader  # noqa
        from pipeline.contrib.external_plugins.models import ExternalPackageSource  # noqa

        # load external components when start command in trigger list
        if self.should_load_external_module():
            try:
                logger.info("Start to update package source from config file...")
                ExternalPackageSource.update_package_source_from_config(
                    getattr(settings, "COMPONENTS_PACKAGE_SOURCES", {})
                )
            except ProgrammingError:
                logger.warning(
                    "update package source failed, maybe first migration? "
                    "exception: {traceback}".format(traceback=traceback.format_exc())
                )
                # first migrate
                return

            logger.info("Start to load external modules...")

            loader.load_external_modules()

    @staticmethod
    def should_load_external_module():
        django_command = env.get_django_command()
        if django_command is None:
            print("app is not start with django manage command, current argv: {argv}".format(argv=sys.argv))
            return True

        triggers = getattr(settings, "EXTERNAL_COMPONENTS_LOAD_TRIGGER", DEFAULT_TRIGGERS)
        print("should_load_external_module: {}".format(django_command in triggers))
        return django_command in triggers



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging
from abc import ABCMeta

import ujson as json

from pipeline.core.constants import ESCAPED_CHARS
from pipeline.core.data.expression import ConstantTemplate, deformat_constant_key
from pipeline.core.flow.base import FlowNode
from pipeline.exceptions import ConditionExhaustedException, EvaluationException, InvalidOperationException
from pipeline.utils.boolrule import BoolRule

logger = logging.getLogger("pipeline_engine")


class Gateway(FlowNode, metaclass=ABCMeta):
    pass


class ExclusiveGateway(Gateway):
    def __init__(self, id, conditions=None, name=None, data=None):
        super(ExclusiveGateway, self).__init__(id, name, data)
        self.conditions = conditions or []

    def add_condition(self, condition):
        self.conditions.append(condition)

    def next(self, data=None):
        default_flow = self.outgoing.default_flow()
        next_flow = self._determine_next_flow_with_boolrule(data)

        if not next_flow:  # determine fail
            if not default_flow:  # try to use default flow
                raise ConditionExhaustedException(
                    "all conditions of branches are False " "while default flow is not appointed"
                )
            return default_flow.target

        return next_flow.target

    def target_for_sequence_flow(self, flow_id):
        flow_to_target = {c.sequence_flow.id: c.sequence_flow.target for c in self.conditions}
        if flow_id not in flow_to_target:
            raise InvalidOperationException("sequence flow(%s) does not exist." % flow_id)
        return flow_to_target[flow_id]

    @staticmethod
    def _transform_escape_char(string):
        """
        对未转义的字符串进行转义，现有的转义字符包括\n, \r, \t
        """
        if not isinstance(string, str):
            return string
        # 已转义的情况
        if len([c for c in ESCAPED_CHARS.values() if c in string]) > 0:
            return string
        for key, value in ESCAPED_CHARS.items():
            if key in string:
                string = string.replace(key, value)
        return string

    def _determine_next_flow_with_boolrule(self, data):
        """
        根据当前传入的数据判断下一个应该流向的 flow （ 不使用 eval 的版本）
        :param data:
        :return:
        """
        for key, value in data.items():
            data[key] = self._transform_escape_char(value)
        logger.info("[{}] ready to resolve conditions: {}".format(self.id, [c.evaluate for c in self.conditions]))
        for condition in self.conditions:
            deformatted_data = {deformat_constant_key(key): value for key, value in list(data.items())}
            try:
                logger.info("[{}] before resolve condition: {}".format(self.id, condition.evaluate))
                resolved_evaluate = ConstantTemplate(condition.evaluate).resolve_data(deformatted_data)
                logger.info("[{}] test {} with data {}".format(self.id, resolved_evaluate, data))
                result = BoolRule(resolved_evaluate).test(data)
                logger.info("[{}] {} test result: {}".format(self.id, resolved_evaluate, result))
            except Exception as e:
                raise EvaluationException(
                    "evaluate[%s] fail with data[%s] message: %s"
                    % (condition.evaluate, json.dumps(deformatted_data), e)
                )
            if result:
                return condition.sequence_flow

        return None

    def skip(self):
        return True


class ParallelGateway(Gateway):
    def __init__(self, id, converge_gateway_id, name=None, data=None):
        super(ParallelGateway, self).__init__(id, name, data)
        self.converge_gateway_id = converge_gateway_id

    def next(self):
        raise InvalidOperationException("can not determine next node for parallel gateway.")


class ConditionalParallelGateway(Gateway):
    def __init__(self, id, converge_gateway_id, conditions=None, name=None, data=None):
        super(ConditionalParallelGateway, self).__init__(id, name, data)
        self.converge_gateway_id = converge_gateway_id
        self.conditions = conditions or []

    def add_condition(self, condition):
        self.conditions.append(condition)

    def targets_meet_condition(self, data):

        targets = []

        logger.info("[{}] ready to resolve conditions: {}".format(self.id, [c.evaluate for c in self.conditions]))
        for condition in self.conditions:
            deformatted_data = {deformat_constant_key(key): value for key, value in list(data.items())}
            try:
                logger.info("[{}] before resolve condition: {}".format(self.id, condition.evaluate))
                resolved_evaluate = ConstantTemplate(condition.evaluate).resolve_data(deformatted_data)
                logger.info("[{}] test {} with data {}".format(self.id, resolved_evaluate, data))
                result = BoolRule(resolved_evaluate).test(data)
                logger.info("[{}] {} test result: {}".format(self.id, resolved_evaluate, result))
            except Exception as e:
                raise EvaluationException(
                    "evaluate[%s] fail with data[%s] message: %s"
                    % (condition.evaluate, json.dumps(deformatted_data), e)
                )
            if result:
                targets.append(condition.sequence_flow.target)

        if not targets:
            raise ConditionExhaustedException("all conditions of branches are False")

        return targets

    def target_for_sequence_flows(self, flow_ids):
        flow_to_target = {c.sequence_flow.id: c.sequence_flow.target for c in self.conditions}
        if not set(flow_ids).issubset(set(flow_to_target.keys())):
            not_exist_flow_ids = set(flow_ids) - set(flow_to_target.keys())
            raise InvalidOperationException(f"sequence flows {not_exist_flow_ids} does not exist.")
        return [flow_to_target[flow_id] for flow_id in flow_ids]

    def next(self):
        raise InvalidOperationException("can not determine next node for conditional parallel gateway.")

    def skip(self):
        return True


class ConvergeGateway(Gateway):
    def next(self):
        return self.outgoing.unique_one().target

    def skip(self):
        raise InvalidOperationException("can not skip conditional converge gateway.")


class Condition(object):
    def __init__(self, evaluate, sequence_flow):
        self.evaluate = evaluate
        self.sequence_flow = sequence_flow



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db.models import F

from bamboo_engine import metrics
from bamboo_engine.eri import Schedule, ScheduleType

from pipeline.eri.models import Schedule as DBSchedule


class ScheduleMixin:
    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_SCHEDULE_WRITE_TIME)
    def set_schedule(self, process_id: int, node_id: str, version: str, schedule_type: ScheduleType) -> Schedule:
        """
        设置 schedule 对象

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        :param schedule_type: 调度类型
        :type schedule_type: ScheduleType
        :return: 调度对象实例
        :rtype: Schedule
        """
        try:
            schedule_model = DBSchedule.objects.get(node_id=node_id, version=version)
        except DBSchedule.DoesNotExist:
            schedule_model = DBSchedule.objects.create(
                process_id=process_id, node_id=node_id, type=schedule_type.value, version=version
            )

        return Schedule(
            id=schedule_model.id,
            type=schedule_type,
            process_id=schedule_model.process_id,
            node_id=schedule_model.node_id,
            finished=schedule_model.finished,
            expired=schedule_model.expired,
            version=schedule_model.version,
            times=schedule_model.schedule_times,
        )

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_SCHEDULE_READ_TIME)
    def get_schedule(self, schedule_id: str) -> Schedule:
        """
        获取 Schedule 对象

        :param schedule_id: 调度实例 ID
        :type schedule_id: str
        :return: Schedule 对象实例
        :rtype: Schedule
        """
        schedule_model = DBSchedule.objects.get(id=schedule_id)

        return Schedule(
            id=schedule_model.id,
            type=ScheduleType(schedule_model.type),
            process_id=schedule_model.process_id,
            node_id=schedule_model.node_id,
            finished=schedule_model.finished,
            expired=schedule_model.expired,
            version=schedule_model.version,
            times=schedule_model.schedule_times,
        )

    def get_schedule_with_node_and_version(self, node_id: str, version: str) -> Schedule:
        """
        通过节点 ID 和执行版本来获取 Scheudle 对象

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 执行版本
        :type version: str
        :return: Schedule 对象
        :rtype: Schedule
        """
        schedule_model = DBSchedule.objects.get(node_id=node_id, version=version)

        return Schedule(
            id=schedule_model.id,
            type=ScheduleType(schedule_model.type),
            process_id=schedule_model.process_id,
            node_id=schedule_model.node_id,
            finished=schedule_model.finished,
            expired=schedule_model.expired,
            version=schedule_model.version,
            times=schedule_model.schedule_times,
        )

    def apply_schedule_lock(self, schedule_id: str) -> bool:
        """
        获取 Schedule 对象的调度锁，返回是否成功获取锁

        :param schedule_id: 调度实例 ID
        :type schedule_id: str
        :return: 是否成功获取锁
        :rtype: bool
        """
        return DBSchedule.objects.filter(id=schedule_id, scheduling=False).update(scheduling=True) == 1

    def release_schedule_lock(self, schedule_id: int):
        """
        释放指定 Schedule 的调度锁

        :param schedule_id: Schedule ID
        :type schedule_id: int
        """
        DBSchedule.objects.filter(id=schedule_id, scheduling=True).update(scheduling=False)

    def expire_schedule(self, schedule_id: int):
        """
        将某个 Schedule 对象标记为已过期

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """
        DBSchedule.objects.filter(id=schedule_id).update(expired=True)

    def finish_schedule(self, schedule_id: int):
        """
        将某个 Schedule 对象标记为已完成

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """
        DBSchedule.objects.filter(id=schedule_id).update(finished=True)

    def add_schedule_times(self, schedule_id: int):
        """
        将某个 Schedule 对象的调度次数 +1

        :param schedule_id: 调度实例 ID
        :type schedule_id: int
        """
        DBSchedule.objects.filter(id=schedule_id).update(schedule_times=F("schedule_times") + 1)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.builder.flow.base import *  # noqa
from pipeline.utils.collections import FancyDict

__all__ = ["ServiceActivity", "SubProcess"]


class ServiceActivity(Element):
    def __init__(
        self,
        component_code=None,
        failure_handler=None,
        error_ignorable=False,
        timeout=None,
        skippable=True,
        retryable=True,
        *args,
        **kwargs
    ):
        self.component = FancyDict({"code": component_code, "inputs": FancyDict({})})
        self.failure_handler = (
            "{module}.{name}".format(module=failure_handler.__module__, name=failure_handler.__name__)
            if failure_handler
            else None
        )
        self.error_ignorable = error_ignorable
        self.timeout = timeout
        self.skippable = skippable
        self.retryable = retryable
        super(ServiceActivity, self).__init__(*args, **kwargs)

    def type(self):
        return PE.ServiceActivity

    def component_dict(self):
        return {
            "code": self.component.code,
            "inputs": {key: var.to_dict() for key, var in list(self.component.inputs.items())},
        }


class SubProcess(Element):
    def __init__(
        self,
        start=None,
        data=None,
        params=None,
        global_outputs=None,
        replace_id=False,
        template_id=None,
        *args,
        **kwargs
    ):
        self.start = start
        self.data = data
        self.params = params or {}
        self.replace_id = replace_id
        self.global_outputs = FancyDict(global_outputs or {})
        self.template_id = template_id
        super(SubProcess, self).__init__(*args, **kwargs)

    def type(self):
        return PE.SubProcess



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import copy
import queue

from pipeline.builder.flow.data import Data, Params
from pipeline.builder.flow.event import ExecutableEndEvent
from pipeline.core.constants import PE
from pipeline.parser.utils import replace_all_id
from pipeline.utils.uniqid import uniqid

__all__ = ["build_tree"]

__skeleton = {
    PE.id: None,
    PE.start_event: None,
    PE.end_event: None,
    PE.activities: {},
    PE.gateways: {},
    PE.flows: {},
    PE.data: {PE.inputs: {}, PE.outputs: {}},
}

__node_type = {
    PE.ServiceActivity: PE.activities,
    PE.SubProcess: PE.activities,
    PE.EmptyEndEvent: PE.end_event,
    PE.EmptyStartEvent: PE.start_event,
    PE.ParallelGateway: PE.gateways,
    PE.ConditionalParallelGateway: PE.gateways,
    PE.ExclusiveGateway: PE.gateways,
    PE.ConvergeGateway: PE.gateways,
}

__start_elem = {PE.EmptyStartEvent}

__end_elem = {PE.EmptyEndEvent}

__multiple_incoming_type = {
    PE.ServiceActivity,
    PE.ConvergeGateway,
    PE.EmptyEndEvent,
    PE.ParallelGateway,
    PE.ConditionalParallelGateway,
    PE.ExclusiveGateway,
    PE.SubProcess,
}

__incoming = "__incoming"


def build_tree(start_elem, id=None, data=None, replace_id=False):
    tree = copy.deepcopy(__skeleton)
    elem_queue = queue.Queue()
    processed_elem = set()

    tree[__incoming] = {}
    elem_queue.put(start_elem)

    while not elem_queue.empty():
        # get elem
        elem = elem_queue.get()

        # update node when we meet again
        if elem.id in processed_elem:
            __update(tree, elem)
            continue

        # add to queue
        for e in elem.outgoing:
            elem_queue.put(e)

        # mark as processed
        processed_elem.add(elem.id)

        # tree grow
        __grow(tree, elem)

    del tree[__incoming]
    tree[PE.id] = id or uniqid()
    user_data = data.to_dict() if isinstance(data, Data) else data
    tree[PE.data] = user_data or tree[PE.data]
    if replace_id:
        replace_all_id(tree)
    return tree


def __update(tree, elem):
    node_type = __node_type[elem.type()]
    node = tree[node_type] if node_type == PE.end_event else tree[node_type][elem.id]
    node[PE.incoming] = tree[__incoming][elem.id]


def __grow(tree, elem):
    if elem.type() in __start_elem:
        outgoing = uniqid()
        tree[PE.start_event] = {
            PE.incoming: "",
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.id: elem.id,
            PE.name: elem.name,
        }

        next_elem = elem.outgoing[0]
        __grow_flow(tree, outgoing, elem, next_elem)

    elif elem.type() in __end_elem or isinstance(elem, ExecutableEndEvent):
        tree[PE.end_event] = {
            PE.incoming: tree[__incoming][elem.id],
            PE.outgoing: "",
            PE.type: elem.type(),
            PE.id: elem.id,
            PE.name: elem.name,
        }

    elif elem.type() == PE.ServiceActivity:
        outgoing = uniqid()

        tree[PE.activities][elem.id] = {
            PE.incoming: tree[__incoming][elem.id],
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.id: elem.id,
            PE.name: elem.name,
            PE.error_ignorable: elem.error_ignorable,
            PE.timeout: elem.timeout,
            PE.skippable: elem.skippable,
            PE.retryable: elem.retryable,
            PE.component: elem.component_dict(),
            PE.optional: False,
            PE.failure_handler: elem.failure_handler,
        }

        next_elem = elem.outgoing[0]
        __grow_flow(tree, outgoing, elem, next_elem)

    elif elem.type() == PE.SubProcess:
        outgoing = uniqid()

        subprocess_param = elem.params.to_dict() if isinstance(elem.params, Params) else elem.params

        subprocess = {
            PE.id: elem.id,
            PE.incoming: tree[__incoming][elem.id],
            PE.name: elem.name,
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.params: subprocess_param,
        }

        if elem.template_id:
            subprocess[PE.template_id] = elem.template_id
        else:
            subprocess[PE.pipeline] = build_tree(
                start_elem=elem.start, id=elem.id, data=elem.data, replace_id=elem.replace_id
            )

        tree[PE.activities][elem.id] = subprocess

        next_elem = elem.outgoing[0]
        __grow_flow(tree, outgoing, elem, next_elem)

    elif elem.type() == PE.ParallelGateway:
        outgoing = [uniqid() for _ in range(len(elem.outgoing))]

        tree[PE.gateways][elem.id] = {
            PE.id: elem.id,
            PE.incoming: tree[__incoming][elem.id],
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.name: elem.name,
        }

        for i, next_elem in enumerate(elem.outgoing):
            __grow_flow(tree, outgoing[i], elem, next_elem)

    elif elem.type() in {PE.ExclusiveGateway, PE.ConditionalParallelGateway}:
        outgoing = [uniqid() for _ in range(len(elem.outgoing))]

        tree[PE.gateways][elem.id] = {
            PE.id: elem.id,
            PE.incoming: tree[__incoming][elem.id],
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.name: elem.name,
            PE.conditions: elem.link_conditions_with(outgoing),
        }

        for i, next_elem in enumerate(elem.outgoing):
            __grow_flow(tree, outgoing[i], elem, next_elem)

    elif elem.type() == PE.ConvergeGateway:
        outgoing = uniqid()

        tree[PE.gateways][elem.id] = {
            PE.id: elem.id,
            PE.incoming: tree[__incoming][elem.id],
            PE.outgoing: outgoing,
            PE.type: elem.type(),
            PE.name: elem.name,
        }

        next_elem = elem.outgoing[0]
        __grow_flow(tree, outgoing, elem, next_elem)

    else:
        raise Exception()


def __grow_flow(tree, outgoing, elem, next_element):
    tree[PE.flows][outgoing] = {PE.is_default: False, PE.source: elem.id, PE.target: next_element.id, PE.id: outgoing}
    if next_element.type() in __multiple_incoming_type:
        tree[__incoming].setdefault(next_element.id, []).append(outgoing)
    else:
        tree[__incoming][next_element.id] = outgoing



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
# -- XXX This module must not use translation as that causes
# -- a recursive loader import!


from django.conf import settings
from django.utils import timezone


def make_aware(value):
    if settings.USE_TZ:
        # naive datetimes are assumed to be in UTC.
        if timezone.is_naive(value):
            value = timezone.make_aware(value, timezone.utc)
        # then convert to the Django configured timezone.
        default_tz = timezone.get_default_timezone()
        value = timezone.localtime(value, default_tz)
    return value


def now():
    return make_aware(timezone.now())



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from pipeline.utils.uniqid import node_uniqid, line_uniqid
from pipeline.core.constants import PE
from pipeline.exceptions import NodeNotExistException

logger = logging.getLogger("root")

BRANCH_SELECT_GATEWAYS = {PE.ExclusiveGateway, PE.ConditionalParallelGateway}


def recursive_replace_id(pipeline_data):
    pipeline_data[PE.id] = node_uniqid()
    replace_all_id(pipeline_data)
    activities = pipeline_data[PE.activities]
    for act_id, act in list(activities.items()):
        if act[PE.type] == PE.SubProcess:
            recursive_replace_id(act[PE.pipeline])
            act[PE.pipeline][PE.id] = act_id


def replace_all_id(pipeline_data):
    flows = pipeline_data[PE.flows]
    node_map = {}
    flow_map = {}

    # step.1 replace nodes id

    # replace events id
    start_event_id = node_uniqid()
    end_event_id = node_uniqid()
    node_map[pipeline_data[PE.start_event][PE.id]] = start_event_id
    node_map[pipeline_data[PE.end_event][PE.id]] = end_event_id

    start_event_id_maps = _replace_event_id(flows, pipeline_data[PE.start_event], start_event_id)
    end_event_id_maps = _replace_event_id(flows, pipeline_data[PE.end_event], end_event_id)

    # replace activities id
    activity_id_maps = {}
    activities = pipeline_data[PE.activities]
    keys = list(activities.keys())
    for old_id in keys:
        substituted_id = node_uniqid()
        node_map[old_id] = substituted_id
        _replace_activity_id(flows, activities, old_id, substituted_id)
        activity_id_maps[old_id] = substituted_id

    # replace gateways id
    gateway_id_maps = {}
    gateways = pipeline_data[PE.gateways]
    keys = list(gateways.keys())
    for old_id in keys:
        substituted_id = node_uniqid()
        node_map[old_id] = substituted_id
        _replace_gateway_id(flows, gateways, old_id, substituted_id)
        gateway_id_maps[old_id] = substituted_id

    # step.2 replace flows id
    flow_id_maps = {}
    keys = list(flows.keys())
    for old_id in keys:
        substituted_id = line_uniqid()
        flow_map[old_id] = substituted_id
        _replace_flow_id(flows, old_id, substituted_id, pipeline_data)
        flow_id_maps[old_id] = substituted_id

    # step.3 replace id in data
    _replace_id_in_data(pipeline_data, node_map)

    # step.4 try to replace front end data
    _replace_front_end_data_id(pipeline_data, node_map, flow_map)

    return {
        PE.start_event: start_event_id_maps,
        PE.end_event: end_event_id_maps,
        PE.activities: activity_id_maps,
        PE.gateways: gateway_id_maps,
        PE.flows: flow_id_maps,
        PE.subprocess_detail: {},
    }


def _replace_id_in_data(pipeline_data, node_map):
    for _, var_info in list(pipeline_data.get(PE.data, {}).get(PE.inputs, {}).items()):
        if PE.source_act in var_info:
            if isinstance(var_info[PE.source_act], str):
                var_info[PE.source_act] = node_map[var_info[PE.source_act]]
            else:
                for source_info in var_info[PE.source_act]:
                    source_info[PE.source_act] = node_map[var_info[PE.source_act]]


def _replace_front_end_data_id(pipeline_data, node_map, flow_map):
    if "line" in pipeline_data:
        for line in pipeline_data["line"]:
            line[PE.id] = flow_map[line[PE.id]]
            line[PE.source][PE.id] = node_map[line[PE.source][PE.id]]
            line[PE.target][PE.id] = node_map[line[PE.target][PE.id]]
    if "location" in pipeline_data:
        for location in pipeline_data["location"]:
            location[PE.id] = node_map[location[PE.id]]
    if "constants" in pipeline_data:
        for key, constant in list(pipeline_data[PE.constants].items()):
            source_info = constant.get("source_info", None)
            if source_info:
                replaced_constant = {}
                for source_step, source_keys in list(source_info.items()):
                    try:
                        replaced_constant[node_map[source_step]] = source_keys
                    except KeyError as e:
                        message = "replace pipeline template id error: %s" % e
                        logger.exception(message)
                        raise NodeNotExistException(message)
                    constant["source_info"] = replaced_constant


def _replace_flow_id(flows, flow_id, substituted_id, pipeline_data):
    flow = flows[flow_id]
    flow[PE.id] = substituted_id

    _replace_flow_in_node(flow[PE.source], pipeline_data, substituted_id, flow_id, PE.outgoing)
    _replace_flow_in_node(flow[PE.target], pipeline_data, substituted_id, flow_id, PE.incoming)

    flows.pop(flow_id)
    flows[substituted_id] = flow


def _replace_flow_in_node(node_id, pipeline_data, substituted_id, flow_id, field):
    if node_id in pipeline_data[PE.activities]:
        node = pipeline_data[PE.activities][node_id]
    elif node_id in pipeline_data[PE.gateways]:
        node = pipeline_data[PE.gateways][node_id]
        if node[PE.type] in BRANCH_SELECT_GATEWAYS and field == PE.outgoing:
            _replace_flow_in_exclusive_gateway_conditions(node, substituted_id, flow_id)
    elif node_id == pipeline_data[PE.start_event][PE.id]:
        node = pipeline_data[PE.start_event]
    elif node_id == pipeline_data[PE.end_event][PE.id]:
        node = pipeline_data[PE.end_event]
    sequence = node[field]
    if isinstance(sequence, list):
        i = sequence.index(flow_id)
        sequence.pop(i)
        sequence.insert(i, substituted_id)
    else:
        node[field] = substituted_id


def _replace_flow_in_exclusive_gateway_conditions(gateway, substituted_id, flow_id):
    conditions = gateway[PE.conditions]
    exist_tag = False
    if flow_id in conditions:
        conditions[substituted_id] = conditions[flow_id]
        conditions.pop(flow_id)
        exist_tag = True
    if gateway.get(PE.default_condition, {}).get("flow_id") == flow_id:
        gateway[PE.default_condition]["flow_id"] = substituted_id
        exist_tag = True
    if exist_tag is False:
        message = (
            f"replace exclusive_gateway_conditions error: "
            f"flow {flow_id} not exists in conditions or default_condition"
        )
        logger.error(message)
        raise NodeNotExistException(message)


def _replace_gateway_id(flows, gateways, gateway_id, substituted_id):
    try:
        gateway = gateways[gateway_id]
        gateway[PE.id] = substituted_id

        if gateway[PE.type] == PE.ConvergeGateway:
            flows[gateway[PE.outgoing]][PE.source] = substituted_id
            for flow_id in gateway[PE.incoming]:
                flows[flow_id][PE.target] = substituted_id
            # replace converge_gateway_id
            for g_id, gw in list(gateways.items()):
                if PE.converge_gateway_id in gw and gw[PE.converge_gateway_id] == gateway_id:
                    gw[PE.converge_gateway_id] = substituted_id
        else:
            incoming = gateway[PE.incoming]

            if isinstance(incoming, list):
                for flow_id in incoming:
                    flows[flow_id][PE.target] = substituted_id
            else:
                flows[gateway[PE.incoming]][PE.target] = substituted_id

            for flow_id in gateway[PE.outgoing]:
                flows[flow_id][PE.source] = substituted_id

        gateways.pop(gateway_id)
        gateways[substituted_id] = gateway
    except KeyError as e:
        message = "replace gateway id error: %s" % e
        logger.exception(message)
        raise NodeNotExistException(message)


def _replace_activity_id(flows, activities, act_id, substituted_id):
    try:
        activity = activities[act_id]
        activity[PE.id] = substituted_id

        incoming = activity[PE.incoming]

        if isinstance(incoming, list):
            for s in incoming:
                flows[s][PE.target] = substituted_id
        else:
            flows[activity[PE.incoming]][PE.target] = substituted_id

        flows[activity[PE.outgoing]][PE.source] = substituted_id

        activities.pop(act_id)
        activities[substituted_id] = activity
    except KeyError as e:
        message = "replace activity id error: %s" % e
        logger.exception(message)
        raise NodeNotExistException(message)


def _replace_event_id(flows, event, substituted_id):
    replace_maps = {}
    try:
        replace_maps[event[PE.id]] = substituted_id
        event[PE.id] = substituted_id
        if event[PE.incoming]:
            if isinstance(event[PE.incoming], list):
                for incoming in event[PE.incoming]:
                    flows[incoming][PE.target] = substituted_id
            else:
                flows[event[PE.incoming]][PE.target] = substituted_id
        else:
            flows[event[PE.outgoing]][PE.source] = substituted_id
    except KeyError as e:
        message = "replace event id error: %s" % e
        logger.exception(message)
        raise NodeNotExistException(message)

    return replace_maps



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-02-13 07:38
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0023_set_is_revoked"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pipelineinstance",
            name="execution_snapshot",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="execution_snapshot_instances",
                to="pipeline.Snapshot",
                verbose_name="用于实例执行的结构数据",
            ),
        ),
        migrations.AlterField(
            model_name="pipelineinstance",
            name="name",
            field=models.CharField(default="default_instance", max_length=128, verbose_name="实例名称"),
        ),
        migrations.AlterField(
            model_name="pipelineinstance",
            name="snapshot",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="snapshot_instances",
                to="pipeline.Snapshot",
                verbose_name="实例结构数据，指向实例对应的模板的结构数据",
            ),
        ),
        migrations.AlterField(
            model_name="pipelineinstance",
            name="tree_info",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="tree_info_instances",
                to="pipeline.TreeInfo",
                verbose_name="提前计算好的一些流程结构数据",
            ),
        ),
        migrations.AlterField(
            model_name="pipelinetemplate",
            name="name",
            field=models.CharField(default="default_template", max_length=128, verbose_name="模板名称"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.conf import settings

# pipeline template context module, to use this, you need
#   1) config PIPELINE_TEMPLATE_CONTEXT in your django settings, such as
#       PIPELINE_TEMPLATE_CONTEXT = 'home_application.utils.get_template_context'
#   2) define get_template_context function in your app, which show accept one arg, such as
#         def get_template_context(obj):
#             context = {
#                 'biz_cc_id': '1',
#                 'biz_cc_name': 'test1',
#             }
#             if obj is not None:
#                 context.update({'template': '1'})
#             return context

PIPELINE_TEMPLATE_CONTEXT = getattr(settings, "PIPELINE_TEMPLATE_CONTEXT", "")
PIPELINE_INSTANCE_CONTEXT = getattr(settings, "PIPELINE_INSTANCE_CONTEXT", "")

PIPELINE_ENGINE_ADAPTER_API = getattr(
    settings, "PIPELINE_ENGINE_ADAPTER_API", "pipeline.service.pipeline_engine_adapter.adapter_api",
)

PIPELINE_DATA_BACKEND = getattr(
    settings, "PIPELINE_DATA_BACKEND", "pipeline.engine.core.data.mysql_backend.MySQLDataBackend",
)
PIPELINE_DATA_CANDIDATE_BACKEND = getattr(settings, "PIPELINE_DATA_CANDIDATE_BACKEND", None)
PIPELINE_DATA_BACKEND_AUTO_EXPIRE = getattr(settings, "PIPELINE_DATA_BACKEND_AUTO_EXPIRE", False)
PIPELINE_DATA_BACKEND_AUTO_EXPIRE_SECONDS = int(
    getattr(settings, "PIPELINE_DATA_BACKEND_AUTO_EXPIRE_SECONDS", 60 * 60 * 24)
)

PIPELINE_END_HANDLER = getattr(
    settings, "PIPELINE_END_HANDLER", "pipeline.engine.signals.handlers.pipeline_end_handler",
)
PIPELINE_WORKER_STATUS_CACHE_EXPIRES = getattr(settings, "PIPELINE_WORKER_STATUS_CACHE_EXPIRES", 30)
PIPELINE_RERUN_MAX_TIMES = getattr(settings, "PIPELINE_RERUN_MAX_TIMES", 0)
PIPELINE_RERUN_INDEX_OFFSET = getattr(settings, "PIPELINE_RERUN_INDEX_OFFSET", -1)

COMPONENT_AUTO_DISCOVER_PATH = [
    "components.collections",
]

COMPONENT_AUTO_DISCOVER_PATH += getattr(settings, "COMPONENT_PATH", [])

AUTO_UPDATE_COMPONENT_MODELS = getattr(settings, "AUTO_UPDATE_COMPONENT_MODELS", True)

VARIABLE_AUTO_DISCOVER_PATH = [
    "variables.collections",
]

VARIABLE_AUTO_DISCOVER_PATH += getattr(settings, "VARIABLE_PATH", [])

AUTO_UPDATE_VARIABLE_MODELS = getattr(settings, "AUTO_UPDATE_VARIABLE_MODELS", True)

PIPELINE_PARSER_CLASS = getattr(settings, "PIPELINE_PARSER_CLASS", "pipeline.parser.pipeline_parser.PipelineParser")

ENABLE_EXAMPLE_COMPONENTS = getattr(settings, "ENABLE_EXAMPLE_COMPONENTS", True)

UUID_DIGIT_STARTS_SENSITIVE = getattr(settings, "UUID_DIGIT_STARTS_SENSITIVE", False)

PIPELINE_LOG_LEVEL = getattr(settings, "PIPELINE_LOG_LEVEL", "INFO")

# 远程插件包源默认配置
EXTERNAL_PLUGINS_SOURCE_PROXY = getattr(settings, "EXTERNAL_PLUGINS_SOURCE_PROXY", None)
EXTERNAL_PLUGINS_SOURCE_SECURE_RESTRICT = getattr(settings, "EXTERNAL_PLUGINS_SOURCE_SECURE_RESTRICT", True)

# 僵尸进程扫描配置
ENGINE_ZOMBIE_PROCESS_DOCTORS = getattr(settings, "ENGINE_ZOMBIE_PROCESS_DOCTORS", None)
ENGINE_ZOMBIE_PROCESS_HEAL_CRON = getattr(settings, "ENGINE_ZOMBIE_PROCESS_HEAL_CRON", {"minute": "*/10"})

# 过期任务运行时清理配置
EXPIRED_TASK_CLEAN = getattr(settings, "EXPIRED_TASK_CLEAN", False)
EXPIRED_TASK_CLEAN_CRON = getattr(settings, "EXPIRED_TASK_CLEAN_CRON", {"minute": "37", "hour": "*"})
EXPIRED_TASK_CLEAN_NUM_LIMIT = getattr(settings, "EXPIRED_TASK_CLEAN_NUM_LIMIT", 100)
TASK_EXPIRED_MONTH = getattr(settings, "TASK_EXPIRED_MONTH", 6)

# MAKO sandbox config
MAKO_SANDBOX_SHIELD_WORDS = getattr(settings, "MAKO_SANDBOX_SHIELD_WORDS", [])
MAKO_SANDBOX_IMPORT_MODULES = getattr(settings, "MAKO_SANDBOX_IMPORT_MODULES", {})
MAKO_SAFETY_CHECK = getattr(settings, "MAKO_SAFETY_CHECK", True)

# 开发者自定义插件和变量异常类
PLUGIN_SPECIFIC_EXCEPTIONS = getattr(settings, "PLUGIN_SPECIFIC_EXCEPTIONS", ())
VARIABLE_SPECIFIC_EXCEPTIONS = getattr(settings, "VARIABLE_SPECIFIC_EXCEPTIONS", ())



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import traceback

from django.utils.module_loading import import_string

from pipeline.conf import settings
from pipeline.core.flow.activity import ServiceActivity
from pipeline.core.pipeline import Pipeline
from pipeline.engine import models, signals
from pipeline.engine.exceptions import InvalidPipelineEndHandleError
from pipeline.engine.signals import handlers

try:
    end_handler = import_string(settings.PIPELINE_END_HANDLER)
except ImportError:
    raise InvalidPipelineEndHandleError(
        "pipeline end handler ({}) import error with exception: {}".format(
            settings.PIPELINE_END_HANDLER, traceback.format_exc()
        )
    )


# DISPATCH_UID = __name__.replace('.', '_')


def dispatch_pipeline_ready():
    signals.pipeline_ready.connect(handlers.pipeline_ready_handler, sender=Pipeline, dispatch_uid="_pipeline_ready")


def dispatch_pipeline_end():
    signals.pipeline_end.connect(end_handler, sender=Pipeline, dispatch_uid="_pipeline_end")


def dispatch_child_process_ready():
    signals.child_process_ready.connect(
        handlers.child_process_ready_handler, sender=models.PipelineProcess, dispatch_uid="_child_process_ready"
    )


def dispatch_process_ready():
    signals.process_ready.connect(
        handlers.process_ready_handler, sender=models.PipelineProcess, dispatch_uid="_process_ready"
    )


def dispatch_batch_process_ready():
    signals.batch_process_ready.connect(
        handlers.batch_process_ready_handler, sender=models.PipelineProcess, dispatch_uid="_batch_process_ready"
    )


def dispatch_wake_from_schedule():
    signals.wake_from_schedule.connect(
        handlers.wake_from_schedule_handler, sender=models.ScheduleService, dispatch_uid="_wake_from_schedule"
    )


def dispatch_schedule_ready():
    signals.schedule_ready.connect(
        handlers.schedule_ready_handler, sender=models.ScheduleService, dispatch_uid="_schedule_ready"
    )


def dispatch_process_unfreeze():
    signals.process_unfreeze.connect(
        handlers.process_unfreeze_handler, sender=models.PipelineProcess, dispatch_uid="_process_unfreeze"
    )


def dispatch_service_activity_timeout_monitor_start():
    signals.service_activity_timeout_monitor_start.connect(
        handlers.service_activity_timeout_monitor_start_handler,
        sender=ServiceActivity,
        dispatch_uid="_service_activity_timeout_monitor_start",
    )


def dispatch_service_activity_timeout_monitor_end():
    signals.service_activity_timeout_monitor_end.connect(
        handlers.service_activity_timeout_monitor_end_handler,
        sender=ServiceActivity,
        dispatch_uid="__service_activity_timeout_monitor_end",
    )


def dispatch():
    dispatch_pipeline_ready()
    dispatch_pipeline_end()
    dispatch_child_process_ready()
    dispatch_process_ready()
    dispatch_batch_process_ready()
    dispatch_wake_from_schedule()
    dispatch_schedule_ready()
    dispatch_process_unfreeze()
    dispatch_service_activity_timeout_monitor_start()
    dispatch_service_activity_timeout_monitor_end()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import urllib.parse

import requests

from pipeline.contrib.external_plugins.utils.importer.base import AutoInstallRequirementsImporter

logger = logging.getLogger("root")


class GitRepoModuleImporter(AutoInstallRequirementsImporter):
    def __init__(self, name, modules, repo_raw_url, branch, use_cache=True, secure_only=True, proxy=None):
        super(GitRepoModuleImporter, self).__init__(name=name, modules=modules)

        if secure_only and not repo_raw_url.startswith("https"):
            raise ValueError("Only accept https when secure_only is True.")
        elif not secure_only:
            logger.warning("Using not secure protocol is extremely dangerous!!")

        self.repo_raw_url = repo_raw_url if repo_raw_url.endswith("/") else "%s/" % repo_raw_url
        self.branch = branch
        self.use_cache = use_cache
        self.file_cache = {}
        self.proxy = proxy or {}

    def is_package(self, fullname):
        return self._fetch_repo_file(self._file_url(fullname, is_pkg=True)) is not None

    def get_code(self, fullname):
        return compile(self.get_source(fullname), self.get_file(fullname), "exec")

    def get_source(self, fullname):
        source_code = self._fetch_repo_file(self._file_url(fullname, is_pkg=self.is_package(fullname)))

        if source_code is None:
            raise ImportError(
                "Can not find {module} in {repo}{branch}".format(
                    module=fullname, repo=self.repo_raw_url, branch=self.branch
                )
            )
        return source_code

    def get_path(self, fullname):
        return [self._file_url(fullname, is_pkg=True).rpartition("/")[0]]

    def get_file(self, fullname):
        return self._file_url(fullname, is_pkg=self.is_package(fullname))

    def _file_url(self, fullname, is_pkg=False):
        base_url = "%s/" % urllib.parse.urljoin(self.repo_raw_url, self.branch)
        path = fullname.replace(".", "/")
        file_name = "%s/__init__.py" % path if is_pkg else "%s.py" % path
        return urllib.parse.urljoin(base_url, file_name)

    def _fetch_repo_file(self, file_url):
        logger.info("Try to fetch git file: {file_url}".format(file_url=file_url))

        if self.use_cache and file_url in self.file_cache:
            logger.info("Use content in cache for git file: {file_url}".format(file_url=file_url))
            return self.file_cache[file_url]

        resp = requests.get(file_url, timeout=10, proxies=self.proxy)

        file_content = resp.content if resp.ok else None

        if self.use_cache:
            self.file_cache[file_url] = file_content
            logger.info("Content cached for git file: {file_url}".format(file_url=file_url))

        return file_content



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import abc
import logging

from django.utils import timezone

from pipeline.core.pipeline import Pipeline
from pipeline.engine import signals, states
from pipeline.engine.models import ProcessCeleryTask, ScheduleService, Status
from pipeline.utils import uniqid

logger = logging.getLogger("celery")


class ZombieProcDoctor(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def confirm(self, proc):
        raise NotImplementedError()

    @abc.abstractmethod
    def cure(self, proc):
        raise NotImplementedError()


class RunningNodeZombieDoctor(ZombieProcDoctor):
    def __init__(self, max_stuck_time: float, detect_wait_callback_proc: bool = False):
        """
        :param max_stuck_time: 最大卡住时间
        :param detect_wait_callback_proc: 是否检测等待回调的进程
        """
        self.max_stuck_time = max_stuck_time
        self.detect_wait_callback_proc = detect_wait_callback_proc

    def confirm(self, proc):

        # do not process none current node
        if not proc.current_node_id:
            logger.warning("Process({}) with current_node({}), skip".format(proc.id, proc.current_node_id))
            return False

        # do not process node status not exist
        try:
            status = Status.objects.get(id=proc.current_node_id)
        except Status.DoesNotExist:
            logger.warning("Process({})'s current_node({}) not exist, skip".format(proc.id, proc.current_node_id))
            return False

        # do not process legacy status data
        if not status.state_refresh_at:
            logger.warning(
                "Process({})'s current_node({}) state_fresh_at({}) is invalid, skip".format(
                    proc.id, proc.current_node_id, status.state_refresh_at
                )
            )
            return False

        # only process RUNNING node
        if status.state != states.RUNNING:
            return False

        try:
            schedule = ScheduleService.objects.schedule_for(status.id, status.version)
        except ScheduleService.DoesNotExist:
            pass
        else:
            if schedule.wait_callback and not self.detect_wait_callback_proc:
                return False

        stuck_time = (timezone.now() - status.state_refresh_at).total_seconds()
        if float(stuck_time) > float(self.max_stuck_time):
            logger.info(
                "Process({}) with current_node({}) stuck_time({}) exceed max_stuck_time({}), "
                "mark as zombie".format(proc.id, proc.current_node_id, stuck_time, self.max_stuck_time)
            )
            return True

        return False

    def cure(self, proc):

        current_node_id = proc.current_node_id

        # try to transit current node to FAILURE
        try:
            result = Status.objects.raw_fail(
                node_id=current_node_id,
                ex_data="This node had been failed because the process diagnode as zombie process",
            )
        except Exception:
            logger.exception(
                "An error occurred when transit node({}) for zombie process({}).".format(current_node_id, proc.id)
            )
        else:
            if not result.result:
                logger.error(
                    "can't not transit node({}) for zombie process({}), message: {}".format(
                        current_node_id, proc.id, result.message
                    )
                )
            else:
                status = result.extra
                status.version = uniqid.uniqid()
                status.save()
                ProcessCeleryTask.objects.revoke(proc.id, kill=True)

                # adjust pipeline state
                proc.adjust_status()
                proc.is_sleep = True
                proc.save()
                logger.info(
                    "Zombie process({}) with node({}) had been cured by {}".format(
                        proc.id, current_node_id, self.__class__.__name__
                    )
                )
        try:
            signals.activity_failed.send(
                sender=Pipeline, pipeline_id=proc.root_pipeline_id, pipeline_activity_id=current_node_id
            )
        except Exception as e:
            logger.exception(
                "An error({}) occurred when send activity_failed signals node({}) "
                "for zombie process({}).".format(e, current_node_id, proc.id)
            )
